#include <EEPROM.h>

/* ---------------------------------------------------------------
  Liefert das Byte von der Position "adresse"
*/
unsigned char getByte(unsigned int adresse)
{
  return(EEPROM.read(adresse));
}

/* ---------------------------------------------------------------
  Schreibt das Byte "wert" an die Position "adresse"
*/
void setByte(unsigned int adresse, unsigned char wert)
{
  EEPROM.write(adresse, wert);
}

/* ---------------------------------------------------------------
  Schreibt das Word "wert" an die Position "adresse"
*/
void setUInt(unsigned int adresse, unsigned int wert)
{
  setByte(adresse, (wert>>8)&0xFF);
  setByte(++adresse, wert&0xFF);
}

/* ---------------------------------------------------------------
  Liefert das Word von der Position "adresse"
*/
unsigned int getUInt(unsigned int adresse)
{
  unsigned int tmp;

  tmp=getByte(adresse++)<<8;
  return(tmp | getByte(adresse));
}

/* ------------------------------------------------------------------------------------
  Kopiert "laenge" Bytes aus "ptr" ins EEPROM ab der Adresse "adr".
  Es werden hinter den Daten noch zwei Bytes Prüfsumme angehängt.
*/
void EEPROMschreibenMitPruefsumme(unsigned int adr, unsigned int laenge, char *ptr)
{
  unsigned int i, checksum;

  for(i=checksum=0; i<laenge; i++) {
    setByte(adr++, *ptr);
    checksum+=*ptr;
    ptr++;
  }
  setUInt(adr, checksum);
}

/* ------------------------------------------------------------------------------------
  Liest "laenge" Bytes ab Adresse "adr" aus dem EEPROM und legt sie ab "ptr" ab.
  Waren die Daten legal, wird 0 zurückgeliefert. Sonst 1.
*/
unsigned char EEPROMlesenMitPruefsumme(unsigned int adr, unsigned int laenge, char *ptr)
{
  unsigned int i, checksum;

  for(i=checksum=0; i<laenge; i++) {
    *ptr=getByte(adr++);
    checksum+=*ptr;
    ptr++;
  }
  if(checksum!=getUInt(adr)) {
    return(1);
  }
  return(0);
}

