/*
    *******************
      Glotz Regulator
    *******************

Code-gesicherte Fernseh-Sperre

Es werden vier Codes verwaltet, mit denen die Stromzufuhr zu einem Fernseher gesteuert wird.
Zwei Codes sind für zwei Kinder. Hier wird je Kind eine tägliche Maximal-Sehdauer eingestellt.
Pro Kind eine Sehdauer für Sonntag bis Donnerstag und eine für Freitag / Samstag.
Weiterhin werden je Kind zwei Zeitbereiche verwaltet, innerhalb derer das TV eingeschaltet
werden kann. Der eine Zeitbereich gilt von Sonntag bis Donnerstag, der andere Freitag und
Samstag.
Zwei weitere Codes sind nicht limitierbar. Einer ist fest programmiert (Mastercode), der
andere ist änderbar. Mit diesen Codes können alle [drei] Codes verändert werden. Nach 
Anmeldung mit einem Kinder-Code, kann nur genau dieser Code geändert werden.
Es kann ein Ferien-Modus aktiviert werden, durch den durchgängig auf die Wochenend-Werte
gestellt werden kann.


  Detlev Ahlgrimm   04.2011
-------------------------------------------
Beschaltung an Arduino UNO mit ATmega328

LCD 4x20
LCD 204B BL (Reichelt) bzw. EA T204B-BNLW mit folgenden Anschlüssen
  1  VSS  GND
  2  VDD  +5V
  3  VEE  Poti (oder zwischen Widerstand 10K an +5V und Diode an GND)
  4  RS   Pin-8
  5  R/W  GND
  6  E    Pin-7
 11  D4   Pin-3 
 12  D5   Pin-4
 13  D6   Pin-5
 14  D7   Pin-6 
 15  LED+ über 1K an Pin-9
 16  LED- GND

DS 1307 (Reichelt)
  1  X1   32.768 MS1V-12,5
  2  X2   32.768 MS1V-12,5
  3  Vbat +3V Bat
  4  GND  GND + (Bat-GND)
  5  SDA  analogPin-4
  6  SCL  analogPin-5
  7  SQR  frei
  8  Vcc  +5V

IR-Empfänger Modul
TSOP 31240 (Reichelt) /  Pins: 1-2---3
  1  GND
  2  +5V
  3  Pin-2

Relais über NPN-Transistor
  1  +5V
  2  Pin-10

PIEZO
  -  GND
  +  Pin-11

Taster
  x  Pin-14 (AnalogInput-0)


------------
v1.5  08.07.2011
- HW auf ATmega328 statt ArduinoUNO + zweites Relais
- HiFi-Power zugefügt
- Fernbedienung erlernbar statt fest auf Samsung
- Timeout-Warnung (eine Minute vorher) nun auch für Zeitfenster

v1.6  12.07.2011
- Anpassungen für "Always on"
--- Power-Off-Countdown wird bei jedem frischen Logon gestartet
--- Tageswechsel wird abgefangen und initialisiert die verbrauchten Zeiten
- Sleep-Timer zugefügt
- MeldungAusgeben() in remoteControlMessung.cpp schickt jetzt bei jedem Meldungskode einen Text ans LCD
- "displayblink==FALSE" aus den else-Zweigen der "eine Minute vor Abschaltung"-Prüfungen entfernt

v1.7 14.07.2011
- "hifi_power_modus_g" wird bei Tageswechsel nicht mehr neu aus NVR geladen. Stattdessen bei Logoff.
- LCD-Beleuchtungs-Steuerung aufgeräumt
- Abfangen illegaler Zeitfenster-Einstellungen
- Anzeige des alten Freischaltkodes bei der Kodeänderung

v1.7a 15.07.2011
- PowerOnOff-Abfrage aus dem Hauptmenü vor den programm_modus-Verteiler verschoben

v1.7b 17.07.2011
- Töne etwas dezenter (Schlafzimmer-tauglich)

v1.7c 23.07.2011
- Modus Parameter eingebaut

v1.8 24.07.2011
- "verfuegbaresZeichenHatMainloopDurchlaufen_g" eingebaut

v1.9 27.12.2011
- laden von "hifi_power_modus_g" in setup()

*/

#include <LiquidCrystal.h>
#include <Wire.h>                       // For some strange reasons, Wire.h must be included here
#include <DS1307new.h>
#include <EEPROM.h>

#include "defines.h"
#include "countdown.h"
#include "helferchen.h"
#include "logging.h"
#include "screen.h"
#include "remoteControlMessung.h"
#include "EEPROMservice.h"

GRcountdown countdown;

//                               1234567890123456
static char menu_titel[][16]= { "Code einstellen",
                                "Zeitfenster    ",
                                "Schalt-Dauern  ",
                                "Datum stellen  ",
                                "Zeit stellen   ",
                                "Log ansehen    ",
                                "Parameter      ",
                                "Ferienmodus    ",
                                "HIFI-Power     "
                              };
static char username[4][5]={ "Tobi", "Lars", "frei", "????" };

byte programm_modus;         // auf einem der Werte aus (#define) MODUS_....

byte eingabeCodeIdx;             // der aktuelle Index in "eingabeCode[]"
char eingabeCode[10];            // Buffer für den Code während der Eingabe als ASCII
unsigned long eingegebenerCode;  // der fertige Code als Zahl

#define WOCHENTAG 0
#define WOCHENENDE 1
#define START 0
#define ENDE 1
#define STUNDE 0
#define MINUTE 1
byte wochen_tag_wochen_ende_g;        // 0=So-Do 1=Fr/Sa
static char wtwe[2][6]={ "So-Do", "Fr/Sa" };
static char wochentage[7][3]={ "So", "Mo", "Di", "Mi", "Do", "Fr", "Sa" };

byte edit_stunde, edit_minute;        // Stunde / Minute zum Editieren
byte edit_stunde_oder_minute;         // 0=Stunde, 1=Minute / welches Element wird gerade editiert
byte edit_tag, edit_monat, edit_jahr; // Tag / Monat / Jaht zum Editieren
byte edit_tag_monat_jahr;             // 0=Tag, 1=Monat, 2=Jahr / welches Element wird gerade editiert
byte zeitfenster_edit_pos;            // 0 - 3 / welches Element im Menü Zeitfenster wird gerade bearbeitet
byte edit_hifi_power_modus;           // 0==aus, 1==jetzt temporär ein, 2==immer mit Tobi, 3==immer mit Lars, 4==immer mit T & L, 5==immer mit TV

byte aktives_menue;                   // 0 bis ANZAHL_MENUEPUNKTE / der aktuell angezeigte Menüpunkt
byte user_fuer_aenderung_g;           // auf (#define) USER_.... / der User, für den gerade die Daten bearbeitet werden

byte aktuelle_minute;                 // für den Countup von dauern_g[]
byte alarm_sekunde;                   // steuert über (RTC.second&1) einen Alarmtom im Sekundentakt
byte lcd_beleuchtung;                 // 0="LCD-Beleuchtung aus", 1="LCD-Beleuchtung an"
byte log_neuster, log_aeltester, log_angezeigt;  // Grenzen und aktueller Eintrag im Log

#define LOGON_PAUSCHALE 10  // Minuten, die jeder Logon erstmal pauschal kostet (um Werbe-Abschaltungen zu unterbinden und somit das TV zu schonen)
byte logon_pauschale;       // Minutenzähler für die ersten 'LOGON_PAUSCHALE' Minuten nach Logon

int sleep_in_minuten;       // Anzahl Minuten, nach denen das TV abgeschaltet wird (-1 für "nicht aktiv")

// Belegung EEPROM
// ab Adresse 512 sind 256 Byte vom Log belegt

// persistente Variablen
#define NVR_FERIENMODUS_BYTE  1
byte ferienmodus_g;                   // 0=keine Ferien, 1=Ferien

#define NVR_TAGE2000_UINT16 2
unsigned int tage2000_g;              // Tage seit 01.01.2000 bzw. RTC.cdn

#define NVR_MINUTEN_VERBRAUCHT_0_BYTE 4
#define NVR_MINUTEN_VERBRAUCHT_1_BYTE 5
byte minuten_verbraucht_g[2];         // [0] für verbrauchte Zeit von Tobi, [1] für verbrauchte Zeit von Lars

#define NVR_CODE_0_ULONG 6
#define NVR_CODE_1_ULONG 10
#define NVR_CODE_2_ULONG 14
unsigned long codes_g[4];             // 0=Tobi, 1=Lars, 2=frei, 3=Mastercode

#define NVR_DAUER00_BYTE 18
#define NVR_DAUER01_BYTE 19
#define NVR_DAUER10_BYTE 20
#define NVR_DAUER11_BYTE 21
byte dauern_g[2][2];                  // [0][0]=Tobi So-Do, [0][1]=Tobi Fr/Sa, [1][0]=Lars So-Do, [1][1]=Lars Fr/Sa

#define NVR_ZEITEN_ARRAY16BYTE 22
byte zeiten_g[2][2][2][2];            // [Tobi(0)/Lars(1)] [Wt(0)/We(1)] [start(0)/ende(1)] [stunde(0)/minute(1)]

#define NVR_HIFI_POWER_MODUS_BYTE 38
byte hifi_power_modus_g;              // 0==aus, (1==jetzt temporär ein), 2==immer mit Tobi, 3==immer mit Lars, 4==immer mit T & L, 5==immer mit TV

#define NVR_FREIER_BEREICH 39

/* ---------------------------------------------------------------
*/
void setup()
{
  unsigned char FB_neu_einlesen, NVR_illegal, u, t, s;

  pinMode(IR_PIN, INPUT);
  digitalWrite(IR_PIN, HIGH);  // Pullup ein

  pinMode(LCD_BELEUCHTUNG,OUTPUT);
  pinMode(PIEZO_PIN,OUTPUT);
  pinMode(RELAIS_PIN,OUTPUT);

  pinMode(FB_DATA_RESET, INPUT);      // AnalogIn-0
  digitalWrite(FB_DATA_RESET, HIGH);  // Pullup ein

  FernbedienungInit();
  scrInit();

  scrCls();
  digitalWrite(LCD_BELEUCHTUNG, 1);
  scrPrint(0, 0, "Glotz-Regulator");
  scrPrint(0, 1, "v1.9  27.12.2011");
  scrUpdateDisplay();
  delay(2000);
  scrCls();

  programm_modus=MODUS_HAUPTSCHIRM;
  aktiver_login_g=USER_KEIN;
  logon_pauschale=0;
  sleep_in_minuten=-1;

  if(digitalRead(FB_DATA_RESET)==LOW) FB_neu_einlesen=1;
  else                                FB_neu_einlesen=0;

  if(EEPROMlesenMitPruefsumme(10, sizeof(FB_parameter), (char *)&FB_parameter)>0 || FB_neu_einlesen==1) {
    // 8 Byte Daten + 2 Byte Checksum
    digitalWrite(LCD_BELEUCHTUNG, 1);
    if(FernbedienungVermessen()>0) {
      scrCls();
      scrPrint(0, 0, "Fehlmessung");
      scrPrint(0, 1, "Neustart noetig!");
      scrUpdateDisplay();
      while(1);
    } else {
      EEPROMschreibenMitPruefsumme(10, sizeof(FB_parameter), (char *)&FB_parameter);
    }
    scrCls();
  }
  if(EEPROMlesenMitPruefsumme(20, sizeof(tastenfunktion), (char *)tastenfunktion)>0 || FB_neu_einlesen==1) {
    // ANZAHL_TASTEN*2 Byte Daten + 2 Byte Checksum
    digitalWrite(LCD_BELEUCHTUNG, 1);
    TastencodesZuweisen();
    EEPROMschreibenMitPruefsumme(20, sizeof(tastenfunktion), (char *)tastenfunktion);
    scrCls();
  }

  RTC.getTime();
  NVR_illegal=FALSE;
  if(RTC.day==1 && RTC.month==1 && RTC.year==2000) {  // frische RTC -> Init
    NVR_illegal=TRUE;
  }
  if(NVR_illegal==FALSE) {
    if(RTC.day==0 && RTC.month==0 && RTC.year==2000) {  // frische RTC -> Init
      NVR_illegal=TRUE;
    }
  }
  if(NVR_illegal==FALSE) {
    RTC.getRAM(NVR_ZEITEN_ARRAY16BYTE, (uint8_t *)&zeiten_g, 16);
    for(u=0; u<2; u++) {
      for(t=0; t<2; t++) {
        for(s=0; s<2; s++) {
          if(zeiten_g[u][t][s][STUNDE]>23 || zeiten_g[u][t][s][MINUTE]>59) {  // unsinnige Zeiten -> Init
            NVR_illegal=TRUE;
          }
        }
      }
    }
  }

  if(NVR_illegal==TRUE) {
    RTC.year=2011;
    RTC.setTime();

    ferienmodus_g=FALSE;
    RTC.setRAM(NVR_FERIENMODUS_BYTE, &ferienmodus_g, sizeof(byte));

    codes_g[CODE_TOBI]=123;
    codes_g[CODE_LARS]=456;
    codes_g[CODE_FREI]=789;
    RTC.setRAM(NVR_CODE_0_ULONG, (uint8_t *)&codes_g, sizeof(unsigned long)*3);

    dauern_g[USER_TOBI][WOCHENTAG] =70;
    dauern_g[USER_TOBI][WOCHENENDE]=90;
    dauern_g[USER_LARS][WOCHENTAG] =65;
    dauern_g[USER_LARS][WOCHENENDE]=75;
    RTC.setRAM(NVR_DAUER00_BYTE, (uint8_t *)&dauern_g, sizeof(byte)*4);

    tage2000_g=RTC.cdn-1;
    RTC.setRAM(NVR_TAGE2000_UINT16, (uint8_t *)&tage2000_g, sizeof(unsigned int));

    minuten_verbraucht_g[USER_TOBI]=0;
    minuten_verbraucht_g[USER_LARS]=0;
    RTC.setRAM(NVR_MINUTEN_VERBRAUCHT_0_BYTE, (uint8_t *)&minuten_verbraucht_g, sizeof(byte)*2);

    zeiten_g[USER_TOBI][WOCHENTAG][START][STUNDE]=19;     zeiten_g[USER_TOBI][WOCHENTAG][START][MINUTE]=0;
    zeiten_g[USER_TOBI][WOCHENTAG][ENDE][STUNDE]=21;      zeiten_g[USER_TOBI][WOCHENTAG][ENDE][MINUTE]=0;      // Tobi Wt 19:00 bis 21:00
    zeiten_g[USER_TOBI][WOCHENENDE][START][STUNDE]=18;    zeiten_g[USER_TOBI][WOCHENENDE][START][MINUTE]=0;
    zeiten_g[USER_TOBI][WOCHENENDE][ENDE][STUNDE]=21;     zeiten_g[USER_TOBI][WOCHENENDE][ENDE][MINUTE]=0;     // Tobi We 18:00 bis 21:00
    zeiten_g[USER_LARS][WOCHENTAG][START][STUNDE]=18;     zeiten_g[USER_LARS][WOCHENTAG][START][MINUTE]=0;
    zeiten_g[USER_LARS][WOCHENTAG][ENDE][STUNDE]=19;      zeiten_g[USER_LARS][WOCHENTAG][ENDE][MINUTE]=30;     // Lars Wt 18:00 bis 19:30
    zeiten_g[USER_LARS][WOCHENENDE][START][STUNDE]=17;    zeiten_g[USER_LARS][WOCHENENDE][START][MINUTE]=30;
    zeiten_g[USER_LARS][WOCHENENDE][ENDE][STUNDE]=20;     zeiten_g[USER_LARS][WOCHENENDE][ENDE][MINUTE]=30;    // Lars We 17:30 bis 20:30
    RTC.setRAM(NVR_ZEITEN_ARRAY16BYTE, (uint8_t *)&zeiten_g, 16);

    hifi_power_modus_g=0;
    RTC.setRAM(NVR_HIFI_POWER_MODUS_BYTE, &hifi_power_modus_g, sizeof(byte));
  }
  RTC.getTime();
  aktuelle_minute=RTC.minute;

  NVR_DatenEinlesen();
  RTC.getRAM(NVR_HIFI_POWER_MODUS_BYTE, &hifi_power_modus_g, sizeof(byte));

  countdown.StartOderRetrigger(RTC.second, COUNTDOWN_LCD_BELEUCHTUNG, 59);
  lcd_beleuchtung=LCD_MENUE_BELEUCHTUNG_EIN;
  countdown.StartOderRetrigger(RTC.second, COUNTDOWN_POWER_ONOFF_SPERRE, 59);

  user_fuer_log_g=USER_KEIN;
/*
  Serial.begin(57600);
  for(u=0; u<16; u++) {
    for(t=0; t<16; t++) {
      Serial.print(EEPROM.read(512+(u*16)+t), HEX);
      Serial.print(" ");
    }
    Serial.println("");
  }
*/

  scrCls();
  isr_ist_aktiv_g=TRUE;
  attachInterrupt(0, Fernbedienung_isr, FALLING);  // Tastatur "aufmachen"
}

/* ---------------------------------------------------------------
  Liest die Daten aus dem NVR ins RAM und Initialisiert bei
  Tageswechsel die Verbrauchs-Zähler der Kinder
*/
void NVR_DatenEinlesen(void)
{
  RTC.getRAM(NVR_FERIENMODUS_BYTE, &ferienmodus_g, sizeof(byte));

  RTC.getRAM(NVR_CODE_0_ULONG, (uint8_t *)&codes_g[CODE_TOBI], sizeof(unsigned long));  // Code Tobi
  RTC.getRAM(NVR_CODE_1_ULONG, (uint8_t *)&codes_g[CODE_LARS], sizeof(unsigned long));  // Code Lars
  RTC.getRAM(NVR_CODE_2_ULONG, (uint8_t *)&codes_g[CODE_FREI], sizeof(unsigned long));  // Code frei
  codes_g[CODE_MSTR]=0xFA18C;  // Mastercode

  RTC.getRAM(NVR_DAUER00_BYTE, &dauern_g[CODE_TOBI][WOCHENTAG],  sizeof(byte));
  RTC.getRAM(NVR_DAUER01_BYTE, &dauern_g[CODE_TOBI][WOCHENENDE], sizeof(byte));
  RTC.getRAM(NVR_DAUER10_BYTE, &dauern_g[CODE_LARS][WOCHENTAG],  sizeof(byte));
  RTC.getRAM(NVR_DAUER11_BYTE, &dauern_g[CODE_LARS][WOCHENENDE], sizeof(byte));

  RTC.getRAM(NVR_TAGE2000_UINT16, (uint8_t *)&tage2000_g, sizeof(unsigned int));
  if(tage2000_g==RTC.cdn) {    // wenn heute schon Zeit verbraucht wurde....
    RTC.getRAM(NVR_MINUTEN_VERBRAUCHT_0_BYTE, (uint8_t *)&minuten_verbraucht_g, sizeof(byte)*2);  // ...dann Verbrauch laden
  } else {
    tage2000_g=RTC.cdn;  // ...sonst Init des Verbrauchs
    RTC.setRAM(NVR_TAGE2000_UINT16, (uint8_t *)&tage2000_g, sizeof(unsigned int));
    minuten_verbraucht_g[USER_TOBI]=0;
    minuten_verbraucht_g[USER_LARS]=0;
    RTC.setRAM(NVR_MINUTEN_VERBRAUCHT_0_BYTE, (uint8_t *)&minuten_verbraucht_g, sizeof(byte)*2);
  }

  RTC.getRAM(NVR_ZEITEN_ARRAY16BYTE, (uint8_t *)&zeiten_g, 16);
  // "hifi_power_modus_g" wird hier nicht initialisiert, damit um 00:00 Uhr nicht plötzlich der DVD-Player aus geht.
  // Der Init bzw. Reload läuft stattdessen bei LogOff.
}


/* ---------------------------------------------------------------
*/
void loop()
{
  // Variablen, die hier angelegt werden, behalten ihren Inhalt nur für den aktuellen Durchlauf von loop() !!!
  byte mrk;
  int i, j;

  if(zeichenVerfuegbar_g==TRUE) {
    verfuegbaresZeichenHatMainloopDurchlaufen_g=TRUE;
  }

  scrUpdateDisplay();
  countdown.Bearbeiten(RTC.second);

  if(countdown.Laeuft(COUNTDOWN_LCD_BELEUCHTUNG)==TRUE) { lcd_beleuchtung |= LCD_MENUE_BELEUCHTUNG_EIN; }
  else                                                  { lcd_beleuchtung &= LCD_MENUE_BELEUCHTUNG_AUS; }

  lcd_beleuchtung &= LCD_BLINK_AUS;
  lcd_beleuchtung |= (RTC.second&1);  // Bit im Sekundentakt togglen

  mrk=0;
  if((lcd_beleuchtung&LCD_MENUE_BELEUCHTUNG_EIN)>0) {
    mrk=1;    // bei LCD_MENUE_BELEUCHTUNG_EIN ist die Beleuchtung unbedingt an
  } else {
    if((lcd_beleuchtung&LCD_BLINK_EIN)>0) {
      if((lcd_beleuchtung&LCD_VOLLER_ALARM_EIN)>0) {
        mrk=1;      // LCD_BLINK_EIN und LCD_VOLLER_ALARM_EIN -> Beleuchtung an
      } else if((lcd_beleuchtung&LCD_STILLER_ALARM_EIN)>0) {
        mrk=1;      // LCD_BLINK_EIN und LCD_STILLER_ALARM_EIN -> Beleuchtung an
      } // else ist mrk=0 (LCD_STILLER_ALARM_AUS und LCD_VOLLER_ALARM_AUS)
    } // else ist mrk=0 (LCD_BLINK_AUS)
  }
  digitalWrite(LCD_BELEUCHTUNG, mrk);

  if((lcd_beleuchtung&LCD_VOLLER_ALARM_EIN)>0) {
    if(alarm_sekunde!=(RTC.second&1)) {
      schlechtTon();  // Logoff-Alarm im Sekunde-Takt (1x zu Beginn der Sekunde)
      alarm_sekunde=(RTC.second&1);
    }
  }

  if(programm_modus>MODUS_CODEEINGABE && countdown.Laeuft(COUNTDOWN_MENUE_OFFEN)==FALSE) {
    // wenn in einem Unter-Menü vom Haputschirm eine Minute keine Taste betätigt wurde, wird zum
    // Hauptschirm zurückgekehrt.
    // ...und es ist ein Feature, dass "ausgetimete" Menüfunktionen wie z.B. "Schaltdauern stellen"
    // die ggf. schon geänderten Werte bis zum nächsten NVR-Daten-Reload bzw. Kaltstart verwenden.
    programm_modus=MODUS_HAUPTSCHIRM;
    scrCls();
  }

  RTC.getTime();
  if(aktuelle_minute!=RTC.minute) {
    aktuelle_minute=RTC.minute;      // wird nur 1x zu Beginn der vollen Minute besucht

    if(tage2000_g!=RTC.cdn) {    // wenn neuer Tag angefangen hat...
      NVR_DatenEinlesen();       // ...dann Init bzw. Reload der NVR-Daten (und somit auch der Verbrauchs-Werte)
    }

    if(sleep_in_minuten>0) { sleep_in_minuten--; }

    if(aktiver_login_g==USER_TOBI || aktiver_login_g==USER_LARS) {  // Zeitverbrauch der Kinder verwalten
      if(logon_pauschale>0) {
        logon_pauschale--;  // erstmal die Logon-Pauschale verbrauchen
      } else {
        minuten_verbraucht_g[aktiver_login_g]++;  // danach dann den jew. Zähler
        RTC.setRAM(NVR_MINUTEN_VERBRAUCHT_0_BYTE, &minuten_verbraucht_g[0], sizeof(byte)*2);
      }
      if(minuten_verbraucht_g[aktiver_login_g]==(maxMinuten(aktiver_login_g)-1))
        { lcd_beleuchtung |= LCD_VOLLER_ALARM_EIN; } // letzte Minute läuft bei TRUE
      if(minuten_verbraucht_g[aktiver_login_g]>=maxMinuten(aktiver_login_g)) {
        lcd_beleuchtung &= LCD_VOLLER_ALARM_AUS;
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_LCD_BELEUCHTUNG, 59);
        aktiver_login_g=USER_KEIN; // Zeit abgelaufen
        sleep_in_minuten=-1;
        writeLog();
      } else {  // Zeitfenter-Logout nur prüfen, wenn die Zeit nicht eh schon abgelaufen ist
        if(istEineMinuteVorZeitfensterEnde(aktiver_login_g)==TRUE)
          { lcd_beleuchtung |= LCD_VOLLER_ALARM_EIN; } // letzte Minute läuft bei TRUE
        if(istInnerhalbZeitfenter(aktiver_login_g)==FALSE) {
          lcd_beleuchtung &= LCD_VOLLER_ALARM_AUS;
          countdown.StartOderRetrigger(RTC.second, COUNTDOWN_LCD_BELEUCHTUNG, 59);
          aktiver_login_g=USER_KEIN;  // Zeitfenter verlassen
          sleep_in_minuten=-1;
          writeLog();
        }
      }
    }
  }

  if(aktiver_login_g==USER_KEIN) {
    digitalWrite(RELAIS_PIN, 0);
    digitalWrite(RELAIS2_PIN, 0);
    if(hifi_power_modus_g==1) {
      // bei direktem Neuanmelden eines Kindes bleibt HiFi-Power ggf. temporär an
      RTC.getRAM(NVR_HIFI_POWER_MODUS_BYTE, &hifi_power_modus_g, sizeof(byte));
    }
  } else {
    digitalWrite(RELAIS_PIN, 1);
    mrk=0;
    switch(hifi_power_modus_g) {
      case 0: mrk=0; break;
      case 1: mrk=1; break;
      case 2: if(aktiver_login_g==USER_TOBI) { mrk=1; } break;
      case 3: if(aktiver_login_g==USER_LARS) { mrk=1; } break;
      case 4: if(aktiver_login_g==USER_TOBI || aktiver_login_g==USER_LARS) { mrk=1; } break;
      case 5: mrk=1; break;
      default: mrk=0;
    }
    digitalWrite(RELAIS2_PIN, mrk);
  }

  if(legalesZeichenVerfuegbar(IRgrp_SLEEP)==TRUE && aktiver_login_g!=USER_KEIN) {
    sleep_in_minuten=120;  // jedes Drücken der Sleep-Taste setzt auf Maximal-Zeit
    gutTon();
    countdown.StartOderRetrigger(RTC.second, COUNTDOWN_LCD_BELEUCHTUNG, 59);
  }
  if(sleep_in_minuten==0) {
    aktiver_login_g=USER_KEIN;  // Sleep-Timer abgelaufen -> Nutzer abmelden == TV ausschalten
    sleep_in_minuten=-1;
    writeLog();
  }
  if(sleep_in_minuten==1) { // in des letzten Minute vor Sleep-Abschaltung blinkt die LCD-Beleuchtung
    lcd_beleuchtung |= LCD_STILLER_ALARM_EIN;
    countdown.Stop(COUNTDOWN_LCD_BELEUCHTUNG); // es soll blinken, auch wenn die LCD-Beleuchtung eigentlich an wäre
  } else {
    lcd_beleuchtung &= LCD_STILLER_ALARM_AUS;
  }

  if(legalesZeichenVerfuegbar(IRgrp_ONOFF)==TRUE) {  // wenn PowerOnOff gedrückt wurde...
    if(aktiver_login_g!=USER_KEIN) {                 // ...und wenn jemand angemeldet ist...
      if(countdown.Laeuft(COUNTDOWN_POWER_ONOFF_SPERRE)==TRUE) {
        schlechtTon(); // ist wohl als PowerOn fürs TV gedacht
      } else {                                       // ...und nicht PowerOff-Sperre läuft
        lcd_beleuchtung &= LCD_VOLLER_ALARM_AUS;     // ...dann alles aus
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_LCD_BELEUCHTUNG, 59);
        writeLog();
        gutTon();
        RTC.setRAM(NVR_MINUTEN_VERBRAUCHT_0_BYTE, &minuten_verbraucht_g[0], sizeof(byte)*2);
        aktiver_login_g=USER_KEIN;
        sleep_in_minuten=-1;
        programm_modus=MODUS_HAUPTSCHIRM;
        delay(1000);  // dem TV etwas Zeit geben, um das PowerOff zu verarbeiten
        scrCls();
      }
    } else {
      schlechtTon(); // niemand angemeldet -> PowerOnOff bei Relais-Aus ist Quatsch
    }
  }

  switch(programm_modus) {
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - Hauptschirm
//   +----------------+
//   |11.06.11  12:34 |
//   |Sa ???? 000/000 |
//   +----------------+
    case MODUS_HAUPTSCHIRM:
      scrPrint(0, 0, RTC.day, 2); scrPrint("."); scrPrint(RTC.month, 2); scrPrint("."); scrPrint(RTC.year-2000, 2);
      if(aktiver_login_g==USER_TOBI || aktiver_login_g==USER_LARS) {
        if(sleep_in_minuten>0) { scrPrint(" s"); } else { scrPrint("  "); }
      } else { scrPrint("  "); }
      scrPrint(RTC.hour, 2); scrPrint(":"); scrPrint(RTC.minute, 2); 
      if(ferienmodus_g==TRUE) { scrPrint(0, 1, "--"); }
      else                    { scrPrint(0, 1, wochentage[RTC.dow]); }
      scrPrint(" "); scrPrint(username[aktiver_login_g]); scrPrint(" ");

      switch(aktiver_login_g) {
        case USER_KEIN: scrPrint(minuten_verbraucht_g[USER_TOBI], 3);  scrPrint("/"); scrPrint(minuten_verbraucht_g[USER_LARS], 3); break;
        case USER_TOBI: scrPrint(minuten_verbraucht_g[USER_TOBI], 3);  scrPrint("/"); scrPrint(maxMinuten(USER_TOBI), 3); break;
        case USER_LARS: scrPrint(minuten_verbraucht_g[USER_LARS], 3);  scrPrint("/"); scrPrint(maxMinuten(USER_LARS), 3); break;
        case USER_FREI:
          if(sleep_in_minuten>0) { scrPrint(sleep_in_minuten, 3); scrPrint("/120"); }
          else                   { scrPrint("999/999"); }
          break;
      }

      if(legalesZeichenVerfuegbar(IRgrp_ZIFFERN + IRgrp_MENU)==TRUE) {
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_LCD_BELEUCHTUNG, 59);
        lcd_beleuchtung &= LCD_VOLLER_ALARM_AUS;
        switch(aktuelle_Tastengruppe_g) {
          case IRgrp_ZIFFERN:    // Codeeingabe bzw. Logon beginnt
            gutTon();
            scrCls();
            countdown.StartOderRetrigger(RTC.second, COUNTDOWN_CODEEINGABE_OFFEN, 10);
            programm_modus=MODUS_CODEEINGABE;
            eingabeCodeIdx=0;
            eingabeCode[eingabeCodeIdx++]=aktuelle_Taste_g+'0';
            eingabeCode[eingabeCodeIdx]='\0';
            break;

          case IRgrp_MENU:
            if(aktiver_login_g==USER_KEIN) {
              schlechtTon();
              break;
            }
            if(countdown.Laeuft(COUNTDOWN_MENUE_OFFEN)==FALSE) {    // wenn eine Minute keine Taste beträtigt wurde, ist das Menü dicht
              schlechtTon();
              KurzeMeldungAnzeigen(MELDUNG_MENUEGESPERRT);
              break;
            }
            scrCls();
            programm_modus=MODUS_MENUE;
            gutTon();
            aktives_menue=0;
            if(aktiver_login_g!=USER_FREI) {
              programm_modus=MODUS_CODE_STELLEN;    // Kinder dürfen nur ihren Code ändern
              eingabeCodeIdx=0;
              eingabeCode[0]='\0';
              user_fuer_aenderung_g=aktiver_login_g;
            }
            break;
        }
      }
      break;
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - Codeeingabe
//   +----------------+
//   |Codeeingabe:    |
//   |    *********   |
//   +----------------+
    case MODUS_CODEEINGABE:
      scrPrint(0, 0, "Codeeingabe:");
      scrSetCursor(3, 1);
      i=strlen(eingabeCode);
      for(j=0; j<i; j++) {
        scrPrint("*");
      }
      scrCursorBlink();

      if(countdown.Laeuft(COUNTDOWN_CODEEINGABE_OFFEN)==FALSE) {  // Zeit abgelaufen
        schlechtTon();
        scrCls();
        programm_modus=MODUS_HAUPTSCHIRM;
        break;
      }

      if(legalesZeichenVerfuegbar(IRgrp_ZIFFERN + IRgrp_ENTER + IRgrp_EXIT)==TRUE) {
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_LCD_BELEUCHTUNG, 59);
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_CODEEINGABE_OFFEN, 10);
        switch(aktuelle_Tastengruppe_g) {
          case IRgrp_ZIFFERN:
            if(eingabeCodeIdx>=9) {
              schlechtTon();    // zu lang
              break;
            }
            eingabeCode[eingabeCodeIdx++]=aktuelle_Taste_g+'0';
            eingabeCode[eingabeCodeIdx]='\0';
            gutTon();
            break;

          case IRgrp_EXIT:
            scrCls();
            programm_modus=MODUS_HAUPTSCHIRM;
            gutTon();
            break;

          case IRgrp_ENTER:
              eingegebenerCode=atol(eingabeCode);
              if(eingegebenerCode==codes_g[CODE_TOBI] || eingegebenerCode==codes_g[CODE_LARS]) {  // Code Tobi oder Lars ?
                if(eingegebenerCode==codes_g[CODE_TOBI]) mrk=USER_TOBI; else mrk=USER_LARS;
                if(user_fuer_log_g==mrk) {                                               // wenn Re-Logon...
                  countdown.StartOderRetrigger(RTC.second, COUNTDOWN_MENUE_OFFEN, 59);   // ...sind Zeiten egal. Einfach nur Menü öffnen
                } else if(istInnerhalbZeitfenter(mrk)==TRUE) {                           // sonst, wenn er zur richtigen Zeit kommt...
                  if(minuten_verbraucht_g[mrk]<(maxMinuten(mrk)-LOGON_PAUSCHALE)) {      // ...und noch Zeit übrig hat...
                    minuten_verbraucht_g[mrk]+=LOGON_PAUSCHALE;                          // ...bekommt er erstmal 'LOGON_PAUSCHALE' Minuten abgezogen
                    RTC.setRAM(NVR_MINUTEN_VERBRAUCHT_0_BYTE, &minuten_verbraucht_g[0], sizeof(byte)*2);
                    logon_pauschale=LOGON_PAUSCHALE;                                     // ...die auf dem zweiten Zähler gleich wieder angerechnet werden
                    aktiver_login_g=mrk;                                                 // ...um dann angemeldet zu sein
                    gutTon();
                    countdown.StartOderRetrigger(RTC.second, COUNTDOWN_MENUE_OFFEN, 59); // Menü öffnen
                    if(user_fuer_log_g!=USER_KEIN) {    // wenn vorher schon jemand (anderes) angeloggt war...
                      writeLog();                       // ...dessen Log-Eintag schreiben
                    } else {
                      countdown.StartOderRetrigger(RTC.second, COUNTDOWN_POWER_ONOFF_SPERRE, 59);  // sonst ist es ein frischer Logon
                    }
                    LogEintragVorbereiten();  // neuen User merken
                  } else { schlechtTon(); KurzeMeldungAnzeigen(MELDUNG_ZEITABGELAUFEN); }
                } else { schlechtTon(); KurzeMeldungAnzeigen(MELDUNG_ZEITSPERRE); }
              } else if(eingegebenerCode==codes_g[CODE_FREI] || eingegebenerCode==codes_g[CODE_MSTR]) {  // Code frei oder Mastercode ?
                aktiver_login_g=USER_FREI;
                gutTon();
                countdown.StartOderRetrigger(RTC.second, COUNTDOWN_MENUE_OFFEN, 59);    // Menü öffnen
                if(user_fuer_log_g!=USER_FREI) {     // wenn nicht Re-Logon...
                  if(user_fuer_log_g!=USER_KEIN) {   // wenn vorher schon jemand (anderes) angeloggt war...
                    writeLog();                      // ...dessen Log-Eintag schreiben
                  } else {
                    countdown.StartOderRetrigger(RTC.second, COUNTDOWN_POWER_ONOFF_SPERRE, 59);  // sonst ist es ein frischer Logon
                  }
                  LogEintragVorbereiten();         // ...und neuen User merken
                }                
/*                if(user_fuer_log_g!=USER_KEIN) {     // wenn vorher schon jemand angeloggt war...
                  if(user_fuer_log_g!=USER_FREI) {   // ...und der keinen Re-Logon gemacht hat
                    writeLog();                      // ...dessen Log-Eintag schreiben
                    LogEintragVorbereiten();         // ...und neuen User merken
                  } // else ist Re-Logon
                } else { // frischer Logon
                  LogEintragVorbereiten();
                  countdown.StartOderRetrigger(RTC.second, COUNTDOWN_POWER_ONOFF_SPERRE, 59);
                }*/
              } else { schlechtTon(); KurzeMeldungAnzeigen(MELDUNG_CODEFALSCH); }
              scrCls();
              programm_modus=MODUS_HAUPTSCHIRM;
            break;
        }
      }
      break;
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - Menü
//   +----------------+
//   |Funktionsauswahl|
//   |Code einstellen |
//   +----------------+
    case MODUS_MENUE:
      scrPrint(0, 0, "Funktionsauswahl");
      scrPrint(0, 1, menu_titel[aktives_menue]);

      if(legalesZeichenVerfuegbar(IRgrp_CSRUPDN + IRgrp_ENTER + IRgrp_EXIT)==TRUE) {
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_LCD_BELEUCHTUNG, 59);
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_MENUE_OFFEN, 59);
        gutTon();
        switch(aktuelle_Tastengruppe_g) {
          case IRgrp_CSRUPDN:            // im Kreis über alle Menü-Punkte
            if(aktuelle_Taste_g==IRkey_CSRUP) { aktives_menue=FensterSub(aktives_menue, 1, 0, ANZAHL_MENUEPUNKTE); }
            if(aktuelle_Taste_g==IRkey_CSRDN) { aktives_menue=FensterAdd(aktives_menue, 1, 0, ANZAHL_MENUEPUNKTE); }
            break;

          case IRgrp_ENTER:            // ein Menü-Punkt wurde ausgewählt, Variablen vorbereiten und ins entspr. Sub-Menü verzweigen
            scrCls();
            switch(aktives_menue+10) {
              case MODUS_CODE_STELLEN: // Menü Code einstellen
                programm_modus=MODUS_CODE_STELLEN;
                user_fuer_aenderung_g=aktiver_login_g;
                eingabeCodeIdx=0;
                ltoa(codes_g[user_fuer_aenderung_g], eingabeCode, 10);
//                eingabeCode[0]='\0';
                break;

              case MODUS_ZEITFENSTER: // Menü Zeitfenster
                programm_modus=MODUS_ZEITFENSTER;
                user_fuer_aenderung_g=USER_TOBI;
                wochen_tag_wochen_ende_g=0;
                zeitfenster_edit_pos=0;
                break;

              case MODUS_SCHALTDAUERN: // Menü Schalt-Dauern
                programm_modus=MODUS_SCHALTDAUERN;
                user_fuer_aenderung_g=USER_TOBI;
                wochen_tag_wochen_ende_g=0;
                break;

              case MODUS_DATUM_STELLEN: // Menü Datum stellen
                programm_modus=MODUS_DATUM_STELLEN;
                edit_tag=RTC.day;
                edit_monat=RTC.month;
                edit_jahr=RTC.year-2000;
                edit_tag_monat_jahr=0;
                break;

              case MODUS_ZEIT_STELLEN: // Menü Zeit stellen
                programm_modus=MODUS_ZEIT_STELLEN;
                edit_stunde=RTC.hour;
                edit_minute=RTC.minute;
                edit_stunde_oder_minute=0;  // 0=stunde 1=minute
                break;

              case MODUS_LOG_ANSEHEN: // Menü Log ansehen
                programm_modus=MODUS_LOG_ANSEHEN;
                log_neuster=findeNeustenLogEintrag();
                log_angezeigt=log_neuster;
                log_aeltester=findeAeltestenLogEintrag();
                break;

              case MODUS_PARAMETER: // Menü Parameter
                programm_modus=MODUS_PARAMETER;
                break;

              case MODUS_FERIENMODUS: // Menü Ferienmodus
                programm_modus=MODUS_FERIENMODUS;
                break;

              case MODUS_HIFI_POWER: // Menü HIFI-Power
                edit_hifi_power_modus=hifi_power_modus_g;
                programm_modus=MODUS_HIFI_POWER;
                break;
            }
            break;

          case IRgrp_EXIT:
            scrCls();
            programm_modus=MODUS_HAUPTSCHIRM;
            break;
        }
      }      
      break;
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - Menü Code einstellen
//   +----------------+
//   |Code einstellen |
//   |Tobi 123456789  |
//   +----------------+
    case MODUS_CODE_STELLEN:
      scrPrint(0, 0, menu_titel[MODUS_CODE_STELLEN-10]);
      scrPrint(0, 1, username[user_fuer_aenderung_g]);
      scrPrint(5, 1, "          ");
      scrPrint(5, 1, eingabeCode);
      if(eingabeCodeIdx==0) scrSetCursor(5, 1);
      scrCursorBlink();

      if(legalesZeichenVerfuegbar(IRgrp_ZIFFERN + IRgrp_CSRUPDN + IRgrp_ENTER + IRgrp_EXIT)==TRUE) {
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_LCD_BELEUCHTUNG, 59);
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_MENUE_OFFEN, 59);
        switch(aktuelle_Tastengruppe_g) {
          case IRgrp_CSRUPDN:
            if(aktiver_login_g!=USER_FREI) {
              schlechtTon();
              break;  // nur Erwachsene dürfen alle Codes ändern
            }
            gutTon();
            if(aktuelle_Taste_g==IRkey_CSRUP) { // User mit variablen Code im Kreis auswählen
              if(user_fuer_aenderung_g==USER_TOBI) { user_fuer_aenderung_g=USER_FREI; }
              else                                 { user_fuer_aenderung_g--; }
            } else {
              if(user_fuer_aenderung_g==USER_FREI) { user_fuer_aenderung_g=USER_TOBI; }
              else                                 { user_fuer_aenderung_g++; }
            }
            eingabeCodeIdx=0;
            ltoa(codes_g[user_fuer_aenderung_g], eingabeCode, 10);
            break;

          case IRgrp_ZIFFERN:
            if(eingabeCodeIdx>=9) {        // maximal 9 Stellen von Ziffern entgegennehmen
              schlechtTon();
              break;
            }
            eingabeCode[eingabeCodeIdx++]=aktuelle_Taste_g+'0';
            eingabeCode[eingabeCodeIdx]='\0';
            gutTon();
            break;

          case IRgrp_ENTER:
            scrCls();
            eingegebenerCode=atol(eingabeCode);  
            j=0;  // hier als Flag für "Code illegal" missbraucht
            for(i=0; i<3; i++) {
              if(i!=user_fuer_aenderung_g) {
                if(eingegebenerCode==codes_g[i]) {    // Codes müssen eindeutig sein
                  schlechtTon();
                  KurzeMeldungAnzeigen(MELDUNG_CODE_ILLEGAL);
                  j=1;
                  break; // for-loop
                }
              }
            }
            if(j==0) {  // wenn der Code eindeutig ist...
              if(eingegebenerCode>100000) {  // ...und mindestens 6 Stellen hat
                codes_g[user_fuer_aenderung_g]=eingegebenerCode;  // ...dann festschreiben
                switch(user_fuer_aenderung_g) {
                  case USER_TOBI: RTC.setRAM(NVR_CODE_0_ULONG, (uint8_t *)&eingegebenerCode, sizeof(unsigned long)); break;
                  case USER_LARS: RTC.setRAM(NVR_CODE_1_ULONG, (uint8_t *)&eingegebenerCode, sizeof(unsigned long)); break;
                  case USER_FREI: RTC.setRAM(NVR_CODE_2_ULONG, (uint8_t *)&eingegebenerCode, sizeof(unsigned long)); break;
                }
                gutTon();
              } else {        // hat nicht mind. 6 Stellen
                schlechtTon();
                KurzeMeldungAnzeigen(MELDUNG_ZUKURZ);
                break;
              }
            }
            if(aktiver_login_g!=USER_FREI) { programm_modus=MODUS_HAUPTSCHIRM; }  // Kinder müssen auf den Hauptschirm zurück
            else                           { programm_modus=MODUS_MENUE; }        // Eltern dürfen im Menü bleiben
            break;

          case IRgrp_EXIT:
            scrCls();
            gutTon();
            if(aktiver_login_g!=USER_FREI) { programm_modus=MODUS_HAUPTSCHIRM; }  // Kinder müssen auf den Hauptschirm zurück
            else { programm_modus=MODUS_MENUE; }
            break;
        }
      }
      break;
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - Menü Zeitfenster
//   +----------------+
//   |Zeitfenster Tobi|
//   |Wt 12:23 - 23:45|
//   +----------------+
    case MODUS_ZEITFENSTER:
      scrPrint(0, 0, menu_titel[MODUS_ZEITFENSTER-10]);
      scrPrint(12, 0, username[user_fuer_aenderung_g]);
      if(wochen_tag_wochen_ende_g==0) { scrPrint(0, 1, "Wt "); }
      else                            { scrPrint(0, 1, "We "); }
      scrPrint(zeiten_g[user_fuer_aenderung_g][wochen_tag_wochen_ende_g][START][STUNDE], 2); scrPrint(":");
      scrPrint(zeiten_g[user_fuer_aenderung_g][wochen_tag_wochen_ende_g][START][MINUTE], 2); scrPrint(" - ");
      scrPrint(zeiten_g[user_fuer_aenderung_g][wochen_tag_wochen_ende_g][ENDE][STUNDE], 2);  scrPrint(":");
      scrPrint(zeiten_g[user_fuer_aenderung_g][wochen_tag_wochen_ende_g][ENDE][MINUTE], 2);
      switch(zeitfenster_edit_pos) {
        case 0: scrSetCursor(4, 1); break;    // Start-Stunde
        case 1: scrSetCursor(7, 1); break;    // Start-Minute
        case 2: scrSetCursor(12, 1); break;   // Ende-Stunde
        case 3: scrSetCursor(15, 1); break;   // Ende-Minute
      }        
      scrCursorBlink();

      if(legalesZeichenVerfuegbar(IRgrp_PRGUPDN + IRgrp_CSRUPDN + IRgrp_CSRLTRT + IRgrp_ENTER + IRgrp_EXIT)==TRUE) {
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_LCD_BELEUCHTUNG, 59);
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_MENUE_OFFEN, 59);
        switch(aktuelle_Tastengruppe_g) {
          case IRgrp_PRGUPDN:
            mrk=zeiten_g[user_fuer_aenderung_g][wochen_tag_wochen_ende_g][(zeitfenster_edit_pos>>1)&1][zeitfenster_edit_pos&1];
            if(aktuelle_Taste_g==IRkey_PRGUP) {
              switch(zeitfenster_edit_pos) {
                case 0: mrk=FensterAdd(mrk, 1, 0, 23); break;
                case 1: mrk=FensterAdd(mrk, 5, 0, 59); break;
                case 2: mrk=FensterAdd(mrk, 1, 0, 23); break;
                case 3: mrk=FensterAdd(mrk, 5, 0, 59); break;
              }
            } else {
              switch(zeitfenster_edit_pos) {
                case 0: mrk=FensterSub(mrk, 1, 0, 23); break;
                case 1: mrk=FensterSub(mrk, 5, 0, 59); break;
                case 2: mrk=FensterSub(mrk, 1, 0, 23); break;
                case 3: mrk=FensterSub(mrk, 5, 0, 59); break;
              }
            }
            if(mrk>0 && mrk<5) mrk=0;  // auf gerade 5 Minuten bringen
            zeiten_g[user_fuer_aenderung_g][wochen_tag_wochen_ende_g][(zeitfenster_edit_pos>>1)&1][zeitfenster_edit_pos&1]=mrk;
            gutTon();
            break;

          case IRgrp_CSRUPDN: // "user_fuer_aenderung_g" und "wochen_tag_wochen_ende_g" im Kreis durchlaufen
            ueberUserWtWe();
            gutTon();
            break;

          case IRgrp_CSRLTRT: // im Kreis durchlaufen: start/stunde -> start/minute -> ende/stunde -> ende/minute -> start/stunde -> ...
            if(aktuelle_Taste_g==IRkey_CSRLT) {  (--zeitfenster_edit_pos)&=3; }
            else                              {  (++zeitfenster_edit_pos)&=3; }
            gutTon();
            break;

          case IRgrp_ENTER:
            mrk=0;
            for(i=0; i<2; i++) {
              for(j=0; j<2; j++) {
                if(zeiten_g[i][j][START][STUNDE]>zeiten_g[i][j][ENDE][STUNDE]) { 
                  user_fuer_aenderung_g=i;
                  wochen_tag_wochen_ende_g=j;
                  zeitfenster_edit_pos=2;
                  mrk=1;
                }
                if(zeiten_g[i][j][START][STUNDE]==zeiten_g[i][j][ENDE][STUNDE] && zeiten_g[i][j][START][MINUTE]>zeiten_g[i][j][ENDE][MINUTE]) { 
                  user_fuer_aenderung_g=i;
                  wochen_tag_wochen_ende_g=j;
                  zeitfenster_edit_pos=3;
                  mrk=1;
                }
              }
            }
            if(mrk==1) {
              schlechtTon();
              KurzeMeldungAnzeigen(MELDUNG_ZEITEN_ILLEGAL);
            } else {
              RTC.setRAM(NVR_ZEITEN_ARRAY16BYTE, (uint8_t *)&zeiten_g, 16);  // Änderungen festschreiben
              scrCls();
              programm_modus=MODUS_MENUE;
              gutTon();
            }
            break;

          case IRgrp_EXIT:
            RTC.getRAM(NVR_ZEITEN_ARRAY16BYTE, (uint8_t *)&zeiten_g, 16);  // Reload der NVR-Daten
            scrCls();
            programm_modus=MODUS_MENUE;
            gutTon();
            break;
        }
      }
      break;
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - Menü Schalt-Dauern
//   +----------------+
//   |Schalt-Dauern   |
//   |Tobi Wt    123  |
//   +----------------+
    case MODUS_SCHALTDAUERN:
      scrPrint(0, 0, menu_titel[MODUS_SCHALTDAUERN-10]);
      scrPrint(0, 1, username[user_fuer_aenderung_g]);
      scrPrint(5, 1, wtwe[wochen_tag_wochen_ende_g]);
      scrPrint(11, 1, dauern_g[user_fuer_aenderung_g][wochen_tag_wochen_ende_g], 3);
      scrSetCursor(13, 1);
      scrCursorBlink();

      if(legalesZeichenVerfuegbar(IRgrp_PRGUPDN + IRgrp_CSRUPDN + IRgrp_ENTER + IRgrp_EXIT)==TRUE) {
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_LCD_BELEUCHTUNG, 59);
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_MENUE_OFFEN, 59);
        gutTon();
        switch(aktuelle_Tastengruppe_g) {
          case IRgrp_PRGUPDN:                    // Werte ändern
            if(aktuelle_Taste_g==IRkey_PRGUP) { dauern_g[user_fuer_aenderung_g][wochen_tag_wochen_ende_g]=FensterAdd(dauern_g[user_fuer_aenderung_g][wochen_tag_wochen_ende_g], 5, 0, 239); }
            else                              { dauern_g[user_fuer_aenderung_g][wochen_tag_wochen_ende_g]=FensterSub(dauern_g[user_fuer_aenderung_g][wochen_tag_wochen_ende_g], 5, 0, 239); }
            break;

          case IRgrp_CSRUPDN:          // "user_fuer_aenderung_g" und "wochen_tag_wochen_ende_g" im Kreis durchlaufen
            ueberUserWtWe();
            break;

          case IRgrp_ENTER:            // Änderungen festschreiben
            RTC.setRAM(NVR_DAUER00_BYTE, &dauern_g[0][0], sizeof(byte)*4);
            scrCls();
            programm_modus=MODUS_MENUE;
            break;

          case IRgrp_EXIT:            // alte (gespeicherte) Werte holen
            RTC.getRAM(NVR_DAUER00_BYTE, &dauern_g[0][0], sizeof(byte)*4);
            scrCls();
            programm_modus=MODUS_MENUE;
            break;
        }
      }
      break;
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - Menü Datum stellen
//   +----------------+
//   |Datum stellen   |
//   |   01.02.11     |
//   +----------------+
    case MODUS_DATUM_STELLEN:
      scrPrint(0, 0, menu_titel[MODUS_DATUM_STELLEN-10]);
      scrPrint(3, 1, edit_tag, 2); scrPrint("."); scrPrint(edit_monat, 2); scrPrint("."); scrPrint(edit_jahr, 2);
      switch(edit_tag_monat_jahr) {
          case 0:  scrSetCursor(4, 1);  break;  // Tag
          case 1:  scrSetCursor(7, 1);  break;  // Monat
          case 2:  scrSetCursor(10, 1); break;  // Jahr
      }
      scrCursorBlink();

      if(legalesZeichenVerfuegbar(IRgrp_PRGUPDN + IRgrp_CSRLTRT + IRgrp_ENTER + IRgrp_EXIT)==TRUE) {
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_LCD_BELEUCHTUNG, 59);
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_MENUE_OFFEN, 59);
        gutTon();
        switch(aktuelle_Tastengruppe_g) {
          case IRgrp_PRGUPDN:
            if(aktuelle_Taste_g==IRkey_PRGUP) {    // Werte ändern
              switch(edit_tag_monat_jahr) {
                case 0: edit_tag=  FensterAdd(edit_tag,   1,  1, 31);   break;
                case 1: edit_monat=FensterAdd(edit_monat, 1,  1, 12);   break;
                case 2: edit_jahr= FensterAdd(edit_jahr,  1, 10, 99);   break;
              }
            } else {
              switch(edit_tag_monat_jahr) {
                case 0: edit_tag=  FensterSub(edit_tag,   1,  1, 31);   break;
                case 1: edit_monat=FensterSub(edit_monat, 1,  1, 12);   break;
                case 2: edit_jahr= FensterSub(edit_jahr,  1, 10, 99);   break;
              }
            }
            break;

          case IRgrp_CSRLTRT:           // zu änderndes Element wählen
            if(aktuelle_Taste_g==IRkey_CSRLT) { if(edit_tag_monat_jahr>0) edit_tag_monat_jahr--; }
            else                              { if(edit_tag_monat_jahr<2) edit_tag_monat_jahr++; }
            break;

          case IRgrp_ENTER:
            RTC.stopClock();            // Änderungen an RTC übergeben
            RTC.day=edit_tag;
            RTC.month=edit_monat;
            RTC.year=edit_jahr+2000;
            RTC.setTime();
            RTC.startClock();
            scrCls();
            programm_modus=MODUS_MENUE;

            RTC.getTime();
            for(i=0; i<255; i+=5) {
              if(getUInt(i)>RTC.cdn) {   // wenn im Log zufünftige Datumswerte gefunden werden...
                EEPROM.write(512+i, 0);  // ...ist das vermurkst -> nullen
                EEPROM.write(512+i+1, 0);
              }
            }
            break;

          case IRgrp_EXIT:              // einfach zurück ohne Übergabe an RTC macht mögl. Änderungen rückgängig
            scrCls();
            programm_modus=MODUS_MENUE;
            break;
        }
      }
      break;
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - Menü Zeit stellen
//   +----------------+
//   |Zeit stellen    |
//   |   12:23        |
//   +----------------+
    case MODUS_ZEIT_STELLEN:
      scrPrint(0, 0, menu_titel[MODUS_ZEIT_STELLEN-10]);
      scrPrint(3, 1, edit_stunde, 2); scrPrint(":");  scrPrint(edit_minute, 2);
      if(edit_stunde_oder_minute==0) { scrSetCursor(4, 1); }  // Stunde
      else                           { scrSetCursor(7, 1); }  // Minute
      scrCursorBlink();

      if(legalesZeichenVerfuegbar(IRgrp_PRGUPDN + IRgrp_CSRLTRT + IRgrp_ENTER + IRgrp_EXIT)==TRUE) {
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_LCD_BELEUCHTUNG, 59);
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_MENUE_OFFEN, 59);
        gutTon();
        switch(aktuelle_Tastengruppe_g) {
          case IRgrp_PRGUPDN:
            if(aktuelle_Taste_g==IRkey_PRGUP) {  // Werte ändern
              if(edit_stunde_oder_minute==0) { edit_stunde=FensterAdd(edit_stunde, 1, 0, 23); }
              else                           { edit_minute=FensterAdd(edit_minute, 1, 0, 59); }
            } else {
              if(edit_stunde_oder_minute==0) { edit_stunde=FensterSub(edit_stunde, 1, 0, 23); }
              else                           { edit_minute=FensterSub(edit_minute, 1, 0, 59); }
            }
            break;

          case IRgrp_CSRLTRT:
            (++edit_stunde_oder_minute)&=1;  // zwischen Stunde und Minute toggle'n
            break;

          case IRgrp_ENTER:
            RTC.stopClock();        // Änderungen an RTC übergeben
            RTC.hour=edit_stunde;
            RTC.minute=edit_minute;
            RTC.second=0;
            RTC.setTime();
            RTC.startClock();
            scrCls();
            programm_modus=MODUS_MENUE;
            break;

          case IRgrp_EXIT:        // einfach zurück ohne Übergabe an RTC macht mögl. Änderungen rückgängig
            scrCls();
            programm_modus=MODUS_MENUE;
            break;
        }
      }
      break;
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - Menü Log ansehen
//   +----------------+
//   |Log ansehen     |
//   |T 0102 1123-1234|    Tobi war am 01.02. von 11:23 bis 12:34 angemeldet
//   +----------------+
    case MODUS_LOG_ANSEHEN:
      scrPrint(0, 0, menu_titel[MODUS_LOG_ANSEHEN-10]);
      scrSetCursor(0, 1);
      if(getUInt(log_angezeigt)==0) {
        scrCls();  // es gibt noch keine Logeinträge
        programm_modus=MODUS_MENUE;
        schlechtTon();
        break;
      }
      holeLogEintrag(log_angezeigt);
      scrPrint((char *)strg_g);

      if(legalesZeichenVerfuegbar(IRgrp_CSRUPDN + IRgrp_EXIT + IRgrp_ENTER)==TRUE) {
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_LCD_BELEUCHTUNG, 59);
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_MENUE_OFFEN, 59);
        switch(aktuelle_Tastengruppe_g) {
          case IRgrp_CSRUPDN:
            if(aktuelle_Taste_g==IRkey_CSRUP) { 
              if(log_angezeigt==log_aeltester) {  // wenn schon beim Ältesten angekommen
                schlechtTon();  // ...gehts nicht weiter
                break;
              }
              log_angezeigt=vorigerLogEintrag(log_angezeigt);  // ...sonst einen zurück
              if(getUInt(log_angezeigt)==0) {  // wenn das aber ein leerer Eintrag ist (weil EEPROM noch nicht rundrum beschrieben)
                log_angezeigt=naechsterLogEintrag(log_angezeigt);  // ...dann wieder einen vor
                schlechtTon();    // ...und "meckern"
                break;
              }
              gutTon();
            } else {
              if(log_angezeigt==log_neuster) {  // wenn schon beim Neusten angekommen
                schlechtTon();  // ...gehts nicht weiter
                break;
              }
              log_angezeigt=naechsterLogEintrag(log_angezeigt);  // ...sonst einen vor
              gutTon();
            }
            break;

          case IRgrp_ENTER:
            gutTon();
            scrCls();
            programm_modus=MODUS_MENUE;
            break;

          case IRgrp_EXIT:
            gutTon();
            scrCls();
            programm_modus=MODUS_MENUE;
            break;
        }
      }
      break;
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - Menü Parameter
//   +----------------+
//   |Parameter       |
//   |xxx xxx xxx xxx |
//   +----------------+
    case MODUS_PARAMETER:
      scrPrint(0, 0, menu_titel[MODUS_PARAMETER-10]);
      scrPrint(0, 1, FB_parameter.isr_startbit, 3);
      scrPrint(" ");
      scrPrint(FB_parameter.isr_stopbit, 3);
      scrPrint(" ");
      legalesZeichenVerfuegbar(IRgrp_EXIT);
      scrPrint(aktuelle_Tastengruppe_g, 3);
      scrPrint(" ");
      scrPrint(aktuelle_Taste_g, 3);

      if(legalesZeichenVerfuegbar(IRgrp_EXIT)==TRUE) {
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_LCD_BELEUCHTUNG, 59);
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_MENUE_OFFEN, 59);
        gutTon();
        switch(aktuelle_Tastengruppe_g) {
          case IRgrp_EXIT:
            scrCls();
            programm_modus=MODUS_MENUE;
            break;
        }
      }
      break;
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - Menü Ferienmodus
//   +----------------+
//   |Ferienmodus     |
//   |ist aus         |
//   +----------------+
    case MODUS_FERIENMODUS:
      scrPrint(0, 0, menu_titel[MODUS_FERIENMODUS-10]);
      if(ferienmodus_g==FALSE) { scrPrint(0, 1, "ist aus"); }
      else                     { scrPrint(0, 1, "ist ein"); }

      if(legalesZeichenVerfuegbar(IRgrp_PRGUPDN + IRgrp_ENTER + IRgrp_EXIT)==TRUE) {
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_LCD_BELEUCHTUNG, 59);
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_MENUE_OFFEN, 59);
        gutTon();
        switch(aktuelle_Tastengruppe_g) {
          case IRgrp_PRGUPDN:
            (++ferienmodus_g)&=1;    // Wert einfach nur toggle'n
            break;

          case IRgrp_ENTER:
            RTC.setRAM(NVR_FERIENMODUS_BYTE, &ferienmodus_g, sizeof(byte));  // Wert festschreiben
            scrCls();
            programm_modus=MODUS_MENUE;
            break;

          case IRgrp_EXIT:
            RTC.getRAM(NVR_FERIENMODUS_BYTE, &ferienmodus_g, sizeof(byte));  // alten (gespeicherten) Wert holen
            scrCls();
            programm_modus=MODUS_MENUE;
            break;
        }
      }
      break;
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - Menü HIFI-Power
//   +----------------+
//   |HIFI-Power      |
//   |ist aus         |
//   +----------------+
    case MODUS_HIFI_POWER:
      scrPrint(0, 0, menu_titel[MODUS_HIFI_POWER-10]);
      switch(edit_hifi_power_modus) {
        case 0:  scrPrint(0, 1, "ist aus         "); break;
        case 1:  scrPrint(0, 1, "ist jetzt ein   "); break;
        case 2:  scrPrint(0, 1, "immer mit Tobi  "); break;
        case 3:  scrPrint(0, 1, "immer mit Lars  "); break;
        case 4:  scrPrint(0, 1, "immer mit T & L "); break;
        case 5:  scrPrint(0, 1, "immer mit Glotze"); break;
      }

      if(legalesZeichenVerfuegbar(IRgrp_PRGUPDN + IRgrp_ENTER + IRgrp_EXIT)==TRUE) {
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_LCD_BELEUCHTUNG, 59);
        countdown.StartOderRetrigger(RTC.second, COUNTDOWN_MENUE_OFFEN, 59);
        gutTon();
        switch(aktuelle_Tastengruppe_g) {
          case IRgrp_PRGUPDN:                    // Werte ändern
            if(aktuelle_Taste_g==IRkey_PRGDN) { edit_hifi_power_modus=FensterAdd(edit_hifi_power_modus, 1, 0, 5); }
            else                              { edit_hifi_power_modus=FensterSub(edit_hifi_power_modus, 1, 0, 5); }
            break;

          case IRgrp_ENTER:            // Änderungen festschreiben
            hifi_power_modus_g=edit_hifi_power_modus;
            if(hifi_power_modus_g==1) {
              edit_hifi_power_modus=0;  // die 1 für "jetzt ein" wird nicht im NV-RAM abgelegt. Stattdessen 0.
            }
            RTC.setRAM(NVR_HIFI_POWER_MODUS_BYTE, &edit_hifi_power_modus, sizeof(byte));
            scrCls();
            programm_modus=MODUS_MENUE;
            break;

          case IRgrp_EXIT:
            scrCls();
            programm_modus=MODUS_MENUE;
            break;
        }
      }
      break;
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
  }

  if(verfuegbaresZeichenHatMainloopDurchlaufen_g==TRUE) {  // wird oben vor der ersten Tastenabfrage auf TRUE gesetzt, wenn dort zeichenVerfuegbar_g==TRUE war
    zeichenVerfuegbar_g=FALSE;
    verfuegbaresZeichenHatMainloopDurchlaufen_g=FALSE;
  }

  if(isr_ist_aktiv_g==FALSE && zeichenVerfuegbar_g==FALSE) {  // ISR wieder aktivieren, wenn sie sich deaktiviert hatte und das Zeichen verarbeitet worden ist (oder illegal war)
    isr_ist_aktiv_g=TRUE;
    attachInterrupt(0, Fernbedienung_isr, FALLING);
  }
  delay(10);
}


/* ---------------------------------------------------------------
  Liefert für das Kind "user" TRUE, wenn die aktuelle Zeit
  innerhalb des jeweils erlaubten Zeitfensters ist
*/
byte istInnerhalbZeitfenter(byte user)
{
  byte wtwe;

  if(!(user==USER_TOBI || user==USER_LARS))
    return(TRUE);    // wenn kein Kind -> auch kein Zeitfenster

  if(ferienmodus_g==TRUE) { wtwe=1; }
  else { if(RTC.dow<5) wtwe=0; else wtwe=1; }

  if(RTC.hour<zeiten_g[user][wtwe][START][STUNDE])        return(FALSE);  // aktStunde < StartStunde

  if(RTC.hour==zeiten_g[user][wtwe][START][STUNDE]) {
    if(RTC.minute<zeiten_g[user][wtwe][START][MINUTE])    return(FALSE);  // aktMinute < StartMinute (bei Stunden gleich)
  }

  if(RTC.hour>zeiten_g[user][wtwe][ENDE][STUNDE])         return(FALSE);  // aktStunde > EndStunde

  if(RTC.hour==zeiten_g[user][wtwe][ENDE][STUNDE]) {
    if(RTC.minute>zeiten_g[user][wtwe][ENDE][MINUTE])     return(FALSE);  // aktMinute > EndMinute (bei Stunden gleich)
  }

  return(TRUE);
}

/* ---------------------------------------------------------------
  Liefert für das Kind "user" TRUE, wenn die aktuelle Zeit
  genau eine Minute vor Zeitfenster-Ende ist
*/
byte istEineMinuteVorZeitfensterEnde(byte user)
{
  byte wtwe;

  if(!(user==USER_TOBI || user==USER_LARS))
    return(FALSE);    // wenn kein Kind -> auch kein Zeitfenster

  if(ferienmodus_g==TRUE) { wtwe=1; }
  else { if(RTC.dow<5) wtwe=0; else wtwe=1; }

  if(RTC.hour==zeiten_g[user][wtwe][ENDE][STUNDE]) {
    if(RTC.minute==zeiten_g[user][wtwe][ENDE][MINUTE])     return(TRUE);  // aktMinute == EndMinute (bei Stunden gleich)
  }

  return(FALSE);
}

/* ---------------------------------------------------------------
  Liefert für das Kind "user" die aktuelle Maximal-Glotz-Dauer
*/
int maxMinuten(byte user)
{
  if(ferienmodus_g==TRUE) {         // wenn Ferien sind....
    return(dauern_g[user][WOCHENENDE]);
  } else {
    if(RTC.dow<5) {
      return(dauern_g[user][WOCHENTAG]);  // in der Woche
    } else {
      return(dauern_g[user][WOCHENENDE]);  // Wochenende
    }
  }
}

/* ---------------------------------------------------------------
  Fehlermeldung gemäß MELDUNGs-Code in "nr" ausgeben
*/
void KurzeMeldungAnzeigen(int nr)
{
  countdown.StartOderRetrigger(RTC.second, COUNTDOWN_LCD_BELEUCHTUNG, 59);
  digitalWrite(LCD_BELEUCHTUNG, 1);  // sofort Licht an
  scrCls();
  switch(nr) {
    //                                            1234567890123456
    case MELDUNG_ZEITSPERRE:      scrPrint(2, 1,   "Zeitsperre !");    break;
    case MELDUNG_ZEITABGELAUFEN:  scrPrint(0, 1, "Zeit abgelaufen!");  break;
    case MELDUNG_CODEFALSCH:      scrPrint(2, 1,   "Code falsch !");   break;
    case MELDUNG_ZUKURZ:          scrPrint(5, 1,      "zu kurz !");    break;
    case MELDUNG_MENUEGESPERRT:   scrPrint(0, 1, "Menue gesperrt!");   break;
    case MELDUNG_CODE_ILLEGAL:    scrPrint(2, 1,   "Code illegal!");   break;
    case MELDUNG_ZEITEN_ILLEGAL:  scrPrint(0, 1, "Zeiten illegal!");   break;
    //                                            1234567890123456
  }
  scrUpdateDisplay();
  delay(2000);
  scrCls();
}

/* ---------------------------------------------------------------
  Durchläuft die oberen Dimensionen von zeiten_g bzw. 
  die beiden Dimensionen von dauern_g im Kreis
  also: Tobi/Wt -> Tobi/We -> Lars/Wt -> Lars/We -> Tobi/Wt -> ...
*/
void ueberUserWtWe(void)
{
  int i;

  i=(user_fuer_aenderung_g*2) | wochen_tag_wochen_ende_g;
  if(aktuelle_Taste_g==IRkey_CSRDN) { i--; }
  else                              { i++; }
  user_fuer_aenderung_g=(i>>1)&1;
  wochen_tag_wochen_ende_g=i&1;
}


