#include <DS1307new.h>
#include <EEPROM.h>
#include "defines.h"
#include "logging.h"

byte strg_g[17];          // ein Text-Buffer für eine aufbereitete Log-Zeile
byte aktiver_login_g;     // auf einem Wert aus (#define) USER_....
byte user_fuer_log_g;     // für diesen User wird das Log geschrieben

byte start_stunde_fuer_log_g;
byte start_minute_fuer_log_g;

/* ---------------------------------------------------------------
  Liefert das Byte von der Position "adresse"
*/
byte getByte(byte adresse)
{
  return(EEPROM.read(adresse+512));
}

/* ---------------------------------------------------------------
  Schreibt das Byte "wert" an die Position "adresse"
*/
void setByte(byte adresse, byte wert)
{
  EEPROM.write(adresse+512, wert);
}

/* ---------------------------------------------------------------
  Schreibt das Word "wert" an die Position "adresse"
*/
void setUInt(byte adresse, unsigned int wert)
{
  setByte(adresse, (wert>>8)&0xFF);
  setByte(++adresse, wert&0xFF);
}

/* ---------------------------------------------------------------
  Liefert das Word von der Position "adresse"
*/
unsigned int getUInt(byte adresse)
{
  unsigned int tmp;

  tmp=getByte(adresse++)<<8;
  return(tmp | getByte(adresse));
}

/* ---------------------------------------------------------------
  Liefert entweder die erste unbeschriebene Adresse im EEPROM
  oder wenn bereits alles beschrieben wurde, die älteste
  Position.
  Das EEPROM muß chronologisch fortlaufend beschrieben sein.
*/
byte findeAeltestenLogEintrag(void)
{
  unsigned int adresse, wert1, wert2;

  wert1=getUInt(0);
  if(wert1==0) {
    return(0);  // leeres EEPROM
  }
  for(adresse=5; adresse<255; adresse+=5) { // 51*5=255
    wert2=getUInt(adresse);
    if(wert2<wert1) {
      return(adresse);
    }
    wert1=wert2;
  }
  return(0); // einmal rum, muss also gerade voll beschrieben sein
}

/* ---------------------------------------------------------------
  Eine Log-Position zurück
*/
byte vorigerLogEintrag(byte adresse)
{ if(adresse>=5) { return(adresse-5); } else { return(250); } }

/* ---------------------------------------------------------------
  Eine Log-Position vore
*/
byte naechsterLogEintrag(byte adresse)
{ if(adresse<250) { return(adresse+5); } else { return(0); } }

/* ---------------------------------------------------------------
  Liefert die Log-Position des zuletzt geschriebenen Log-Eintrages
*/
byte findeNeustenLogEintrag(void)
{ return(vorigerLogEintrag(findeAeltestenLogEintrag())); }

/* ---------------------------------------------------------------
  Einen Log-Eintag ins EEPROM schreiben.
  Es wird der älteste Eintrag addressiert und überschrieben.
  "user_fuer_log_g" wird auf "USER_KEIN" gesetzt.
*/
void writeLog(void)
{
  byte adresse;
// zwei Byte Datum bzw. "cdn" und drei Bytes für:
//                 2          1          0
//              32109876 54321098 76543210
// user        uu      |        |         
// startstunde   hhhhh |        |         
// startminute        m|mmmmm   |         
// endstunde           |     hhh|hh      
// endminute           |        |  mmmmmm
//             11111111 22222222 33333333

  adresse=findeAeltestenLogEintrag();

  setUInt(adresse, RTC.cdn);  adresse+=2;
  setByte(adresse++, ((user_fuer_log_g&3)<<6) | ((start_stunde_fuer_log_g&0x1F)<<1) | ((start_minute_fuer_log_g&0x20)>>5) );
  setByte(adresse++, ((start_minute_fuer_log_g&0x1F)<<3) | ((RTC.hour&0x1C)>>2));
  setByte(adresse, ((RTC.hour&3)<<6) | (RTC.minute&0x3F));

  user_fuer_log_g=USER_KEIN;  // gemerkte Startzeit ungültig machen
}

/* ---------------------------------------------------------------
  itoa für bytes < 100
  immer zwei Zeichen, ggf.mit führender '0' und ohne \0
*/
void myitoa(byte wert, byte *ptr)
{
  *ptr=(wert/10)+'0';  ptr++;
  *ptr=(wert%10)+'0';
}

/* ---------------------------------------------------------------
  Den Log-Eintrag zu "adresse" holen und formatiert in strg_g[] ablegen
*/
void holeLogEintrag(byte adresse)
{
  byte b1, b2, b3;
  unsigned int datum_cdn;

  datum_cdn=getUInt(adresse);
  adresse+=2;

  b1=getByte(adresse++);
  b2=getByte(adresse++);
  b3=getByte(adresse);

  switch((b1>>6)&3) {
    case USER_TOBI: strg_g[0]='T'; break;
    case USER_LARS: strg_g[0]='L'; break;
    case USER_FREI: strg_g[0]='E'; break;
    default: strg_g[0]='?';
  }
  strg_g[1]=' ';
  RTC.fillByCDN(datum_cdn);

  myitoa(RTC.day, &strg_g[2]);
  myitoa(RTC.month, &strg_g[4]);
  strg_g[6]=' ';
  myitoa((b1>>1)&0x1F, &strg_g[7]);
  myitoa(((b1&1)<<5) | ((b2>>3)&0x1F), &strg_g[9]);
  strg_g[11]='-';
  myitoa(((b2&7)<<2) | ((b3&0xC0)>>6), &strg_g[12]);
  myitoa((b3&0x3F), &strg_g[14]);
  strg_g[16]='\0';
  RTC.getTime();  // RTC-Variablen wieder auf akt. Datum/Zeit setzen
}

/* ---------------------------------------------------------------
  Start-Daten für Log-Eintrag merken
*/
void LogEintragVorbereiten(void)
{
  user_fuer_log_g=aktiver_login_g;
  start_stunde_fuer_log_g=RTC.hour;
  start_minute_fuer_log_g=RTC.minute;
}

