#include "remoteControlMessung.h"
#include "defines.h"
#include "helferchen.h"
#include "screen.h"

#include <LiquidCrystal.h>
extern LiquidCrystal lcd;


#define SIGNAL_DELTA 15
#define MESS_ISR_TIMEOUT 15000

#define FLANKE_2H 0
#define FLANKE_2L 1

#define ISR_MODE_EINMESSUNG  0
#define ISR_MODE_CODEMESSUNG 1

struct minmax_t {    // 162 Bytes
  unsigned char signalZeiten_idx[2];          // Index, steht immer auf dem nächst freien Platz
  unsigned int  signalZeiten[2][10];          // 2 Flanken, max. 10 Zeiten
  unsigned int  signalZeitenMin[2][10];       // Kleinster Wert zu signalZeiten[][]
  unsigned int  signalZeitenMax[2][10];       // Größter   Wert zu signalZeiten[][]
  unsigned int  signalZeitenAnzahlen[2][10];  // die Auftrittshäufigkeiten zu signalZeiten[][]
};
struct mask_t {      // 200 Bytes
  unsigned char signal[2][50];     // 2 Signal-Infos zu je 50(*2) Bit (4 Bit je Signal)
  unsigned char maske[50];
  unsigned char maske_syncbits[50];
};
union ubuf {
  minmax_t m1;
  mask_t   m2;
};

struct daten_t {    // 230 Bytes, Variablen für die Bit-Dauer-Messung
  unsigned char signalZeiten_idx;   // Index, steht immer auf dem nächst freien Platz
  unsigned int  signalZeiten[10];   // 2 Flanken, max. 10 Zeiten
  unsigned char datenFlanke;        // auf welcher Flanke kommen die Daten
  ubuf          rs;                 // RAM-Sparer
  unsigned char signal_idx;         // index für die Signale - signal[signal_idx][]
  unsigned char signal_sub_idx;     // index innerhalb eines Signals - signal[][signal_sub_idx]
  unsigned char anzahl_tasten;      // Anzahl gezählter Tasten
  unsigned char start_bitpos;       // Position, ab der der Tastencode anfängt
  unsigned char stop_bitpos;        // Position, an der der Tastencode aufhört
  unsigned char signal_prefix_lng;  // die Anzahl der Bits in signal_prefix
  unsigned int  signal_prefix;      // die ersten (bis zu) 16 Bit des Signals zur Identifizierung
};

daten_t ird;
volatile unsigned char zeichenVerfuegbar_g;
volatile unsigned char isr_ist_aktiv_g;
volatile unsigned int isr_tastencode;
unsigned char isr_modus;
unsigned char verfuegbaresZeichenHatMainloopDurchlaufen_g;  // auf TRUE, wenn "zeichenVerfuegbar_g" ab Mail-Loop-Start TRUE war

FB_parameter_t FB_parameter;

// hier stehen die Tastenkodes entsprechend ihrer Funktion
// an den Positionen gemäß "tastennamen[][]".
unsigned int tastenfunktion[ANZAHL_TASTEN];

unsigned int aktuelle_Tastengruppe_g;       // auf einem Wert aus (#define) IRgrp_.... / die Gruppe zu "aktuelle_Taste_g"
unsigned char aktuelle_Taste_g;             // auf einem Wert aus (#define) IRkey_....

/* ------------------------------------------------------------------------------------
  Liefert true, wenn "z" und "v" quasi gleich sind. Quasi gleich sind sie, wenn "z" 
  entweder im Intervall {"v"-SIGNAL_DELTA --- "v"+SIGNAL_DELTA}
  oder im Intervall {"v"-5%"z" --- "v"+5%"z"} liegt.
  Es wird das jeweils breitere Intervall verwendet.
*/
unsigned int istImDeltaBereich(int z, int v)
{
  unsigned int d=SIGNAL_DELTA;

  if((z/20)>d) d=z/20;
  return(z>(v-d) && z<(v+d));
}

/* ------------------------------------------------------------------------------------
*/
void MeldungAusgeben(unsigned char nr)
{
  scrCls();
  switch(nr) {
// Aufgaben
    case 10:
      scrPrint(0, 0, "bitte 5 Tasten");
      scrPrint(0, 1, "druecken");
      break;

    case 11:
      scrPrint(0, 0, "ein paar Tasten");
      scrPrint(0, 1, "lange druecken");
      break;

    case 12:
      scrPrint(0, 0, "EINE Taste 2x");
      scrPrint(0, 1, "LANGE druecken");
      break;

    case 13:
      scrPrint(0, 0, "jede Taste 1x");
      scrPrint(0, 1, "druecken");
      break;

// Fehler
    case 129:
      scrPrint(0, 0, "Fehler 129");
      //Serial.println("Fehler!!! mehr als 10 Zeiten!");
      break;

    case 130:
      scrPrint(0, 0, "Fehler 130");
      //Serial.println("Fehler!!! Kann Datenflanke nicht ermitteln! Reset noetig!");
      break;

// Infos
    case 200:
      scrPrint(0, 0, "Info 200");
      //Serial.println("Probleme mit der SyncBit-Erkennung!");
      break;

    case 201:
      scrPrint(0, 0, "Info 201");
      //Serial.println("Fertig mit der SyncBit-Erkennung!");
      break;

    case 202:
      scrPrint(0, 0, "Info 202");
      //Serial.println("Fertig mit der Codeanalyse-Vorbereitung!");
      break;
    
    default:
      scrPrint(0, 0, "Undefinierte");
      scrPrint(0, 1, "   Meldung");
      //Serial.println("Unbekannte Meldung....!?");
  }
  scrUpdateDisplay();
}

/* ------------------------------------------------------------------------------------
  Liefert den Bitwert an der Position "pos" aus "ird.rs.m2.maske[]".
*/
unsigned char BitInMaskeTesten(unsigned char p)  // p von 0-99
{
  if((p&1)==0) return(ird.rs.m2.maske[p>>1]&0xf);
  return((ird.rs.m2.maske[p>>1]&0xf0)>>4);
}

/* ------------------------------------------------------------------------------------
  Setzt das Bit an der Position "pos" aus "ird.rs.m2.maske[]" auf den Wert "v".
*/
void BitInMaskeSetzen(unsigned char p, unsigned char v)  // p von 0-99
{
  if((p&1)==0) ird.rs.m2.maske[p>>1]|=(v&0xf);
  else         ird.rs.m2.maske[p>>1]|=((v<<4)&0xf0);
}
/* ------------------------------------------------------------------------------------
  Liefert den Bitwert an der Position "pos" aus "ird.rs.m2.maske_syncbits[]".
*/
unsigned char BitInMaskeSyncBitsTesten(unsigned char p)  // p von 0-99
{
  if((p&1)==0) return(ird.rs.m2.maske_syncbits[p>>1]&0xf);
  return((ird.rs.m2.maske_syncbits[p>>1]&0xf0)>>4);
}
/* ------------------------------------------------------------------------------------
  Setzt das Bit an der Position "pos" aus "ird.rs.m2.maske_syncbits[]" auf den Wert "v".
*/
void BitInMaskeSyncBitsSetzen(unsigned char p, unsigned char v)  // p von 0-99
{
  if((p&1)==0) ird.rs.m2.maske_syncbits[p>>1]|=(v&0xf);
  else         ird.rs.m2.maske_syncbits[p>>1]|=((v<<4)&0xf0);
}
/* ------------------------------------------------------------------------------------
  Liefert den Bitwert an der Position "pos" aus "ird.rs.m2.signal[idx][]".
*/
unsigned char BitInSignalTesten(unsigned char idx, unsigned char p)  // p von 0-99
{
  if((p&1)==0) return(ird.rs.m2.signal[idx][p>>1]&0xf);
  return((ird.rs.m2.signal[idx][p>>1]&0xf0)>>4);
}
/* ------------------------------------------------------------------------------------
  Setzt das Bit an der Position "pos" aus "ird.rs.m2.signal[ídx][]" auf den Wert "v".
*/
void BitInSignalSetzen(unsigned char idx, unsigned char p, unsigned char v)  // p von 0-99
{
  if((p&1)==0) ird.rs.m2.signal[idx][p>>1]|=(v&0xf);
  else         ird.rs.m2.signal[idx][p>>1]|=((v<<4)&0xf0);
}

/* ------------------------------------------------------------------------------------
  Löscht eines der 50-Byte-Arrays schnell.
*/
void BufferSchnellLoeschen(unsigned char *ptr)
{
  for(unsigned char i=0; i<50; i++) *ptr++=0;
}

/* ------------------------------------------------------------------------------------
  Liefert die Anzahl der für "b" benötigten Bits.
*/
unsigned char benoetigteBits(unsigned char b)
{
  unsigned char i=1;

  while((b>>=1)>0)  i++;
  return(i);
}

/* ------------------------------------------------------------------------------------
  Liefert den Index des größten Wertes aus "ird.signalZeitenAnzahlen[]".
*/
unsigned char MaxAnzahlIndex(unsigned char flanke)
{
  unsigned char i, amax_idx;
  unsigned int amax;

  for(i=amax=0; i<ird.rs.m1.signalZeiten_idx[flanke]; i++) {
    if(ird.rs.m1.signalZeitenAnzahlen[flanke][i]>amax) {
      amax=ird.rs.m1.signalZeitenAnzahlen[flanke][i];
      amax_idx=i;
    }
  }
  return(amax_idx);
}

/* ------------------------------------------------------------------------------------
  Sortiert das Array "ird.rs.m1.signalZeiten[flanke][]". Das Sortierkriterium ist der
  zugehörige Wert in "ird.signalZeitenAnzahlen[flanke][]".
  "ird.signalZeitenAnzahlen[flanke][]" wird analog mitsortiert.
*/
void signalZeitenSortieren(unsigned char flanke)
{
  unsigned char i, j;
  unsigned int srt_z[10], srt_a[10];

  for(i=0; i<ird.rs.m1.signalZeiten_idx[flanke]; i++) {
    j=MaxAnzahlIndex(flanke);

    srt_z[i]=ird.rs.m1.signalZeiten[flanke][j];
    srt_a[i]=ird.rs.m1.signalZeitenAnzahlen[flanke][j];
    ird.rs.m1.signalZeitenAnzahlen[flanke][j]=0;
  }
  for(i=0; i<ird.rs.m1.signalZeiten_idx[flanke]; i++) {
    ird.rs.m1.signalZeiten[flanke][i]=srt_z[i];
    ird.rs.m1.signalZeitenAnzahlen[flanke][i]=srt_a[i];
  }
}

/* ------------------------------------------------------------------------------------
  Fügt die Zeit "z" einmalig in das Array "ird.rs.m1.signalZeiten[flanke][]" ein.
  Pro Wert werden auch die Minimal- und Maximal-Zeiten vermerkt und der Wert für
  "z" entsprechend mittig korrigiert.
  Weiterhin werden die Häufikeiten pro Wert gespeichert.
  Liefert 0, wenn alles geklappt hat. Sonst 1.
*/
unsigned char zeitEinfuegen(unsigned int z, unsigned char flanke)
{
  unsigned char i;

  if(z==0 || z>=MESS_ISR_TIMEOUT) {
    return(0);
  }

  for(i=0; i<ird.rs.m1.signalZeiten_idx[flanke]; i++) {
    if(istImDeltaBereich(z, ird.rs.m1.signalZeiten[flanke][i])) {
      // "z" ist schon drin
      ird.rs.m1.signalZeitenAnzahlen[flanke][i]++;
      if(z<ird.rs.m1.signalZeitenMin[flanke][i])  ird.rs.m1.signalZeitenMin[flanke][i]=z;
      if(z>ird.rs.m1.signalZeitenMax[flanke][i])  ird.rs.m1.signalZeitenMax[flanke][i]=z;
      // wert = (max-min)/2 + min
      ird.rs.m1.signalZeiten[flanke][i]=((ird.rs.m1.signalZeitenMax[flanke][i]-ird.rs.m1.signalZeitenMin[flanke][i])/2)+ird.rs.m1.signalZeitenMin[flanke][i];
      return(0);
    }
  }
  if(ird.rs.m1.signalZeiten_idx[flanke]>=9) {
    MeldungAusgeben(129); // Fehler!!! mehr als 10 Zeiten!
    return(1);
  }
  // hier gilt: i == ird.rs.m1.signalZeiten_idx[flanke]
  ird.rs.m1.signalZeitenAnzahlen[flanke][i]++;
  ird.rs.m1.signalZeiten[flanke][i]=z;
  ird.rs.m1.signalZeitenMin[flanke][i]=z;
  ird.rs.m1.signalZeitenMax[flanke][i]=z;
  ird.rs.m1.signalZeiten_idx[flanke]++;
  return(0);
}

/* ------------------------------------------------------------------------------------
  Liefert die Nummer der Flanke, die die Nutzdaten enthält. Oder 255 bei Fehler.
*/
unsigned char DatenFlankeFinden(void)
{
  unsigned char i, flanke;
  unsigned int sum, p1, p2;

  for(flanke=0; flanke<2; flanke++) {
    // Test 1: Anzahl der Infos pro Flanke
    if(ird.rs.m1.signalZeiten_idx[flanke]<3) {
      if(ird.rs.m1.signalZeiten_idx[flanke^1]>=3) {
        // es werden in der Datenflanke mindestens die drei Infos "0", "1" und "Startbit" benötigt.
        // Somit steht die Datenflanke fest, wenn die eine <3 und die andere >=3 Infos enthält.
        return(flanke^1);
      }
    }
    // ab hier: Flanke mit mindestens drei Infos
    // Test 2: es sollte zwei Werte mit hoher Anzahl geben (für "0" und "1")
    //         die jew. mindestens 10% und zusammen >=80% der Infos ausmachen.
    for(i=sum=0; i<ird.rs.m1.signalZeiten_idx[flanke]; i++) {
      // Summe über alle Anzahlen bilden
      sum+=ird.rs.m1.signalZeitenAnzahlen[flanke][i];
    }
    p1=(ird.rs.m1.signalZeitenAnzahlen[flanke][0]*10)/(sum/10);
    p2=(ird.rs.m1.signalZeitenAnzahlen[flanke][1]*10)/(sum/10);
    if((p1>=10) && (p2>=10) && ((p1+p2)>=80)) {
      return(flanke);
    }
  }
  return(255);
}

/* ------------------------------------------------------------------------------------
  InterruptServiceRoutine zum Auslesen der Signale eines IR-Empfangsmoduls.
 
  ISR_MODE_EINMESSUNG:
  Erfassen der Signal-Zeiten und Signal-Häufigkeiten.
 
  ISR_MODE_CODEMESSUNG:
  Erfassen der restlichen Attribute von "FB_parameter_t".
*/
void isr_ir(void)  // wird bei IR_PIN=="HIGH->LOW" aufgerufen
{
  unsigned char i, j;
  unsigned int cnt;

  if(digitalRead(IR_PIN)==HIGH) return;

  detachInterrupt(0);
  isr_ist_aktiv_g=0;
  ird.signal_sub_idx=0;

  for(i=0; i<100; i++) {
    cnt=0;
    while(digitalRead(IR_PIN)==LOW) { // LOW-HIGH-Flanke vermessen
      delayMicroseconds(10);
      cnt++;
      if(cnt>=MESS_ISR_TIMEOUT) break;
    }
    switch(isr_modus) {
      case ISR_MODE_EINMESSUNG:
        zeitEinfuegen(cnt, FLANKE_2H);
        break;
  
      case ISR_MODE_CODEMESSUNG:
        if(ird.datenFlanke==FLANKE_2H) {
          for(j=0; j<ird.signalZeiten_idx; j++) {
            if(istImDeltaBereich(cnt, ird.signalZeiten[j])) {
              BitInSignalSetzen(ird.signal_idx, ird.signal_sub_idx, j);
              ird.signal_sub_idx++;
              break;
            }
          }
        }
        break;
    }

    cnt=0;
    while(digitalRead(IR_PIN)==HIGH) { // HIGH-LOW-Flanke vermessen
      delayMicroseconds(10);
      cnt++;
      if(cnt>=MESS_ISR_TIMEOUT) break;
    }
    switch(isr_modus) {
      case ISR_MODE_EINMESSUNG:
        zeitEinfuegen(cnt, FLANKE_2L);
        break;
  
      case ISR_MODE_CODEMESSUNG:
        if(ird.datenFlanke==FLANKE_2L) {
          for(j=0; j<ird.signalZeiten_idx; j++) {
            if(istImDeltaBereich(cnt, ird.signalZeiten[j])) {
              BitInSignalSetzen(ird.signal_idx, ird.signal_sub_idx, j);
              ird.signal_sub_idx++;
              break;
            }
          }
        }
        break;
    }

    if(cnt>=MESS_ISR_TIMEOUT) break;
  }
  zeichenVerfuegbar_g=1;

  if(cnt<MESS_ISR_TIMEOUT) { // Rest ggf. weglesen
    while(1) { // Ende über return
      cnt=0;
      while(digitalRead(IR_PIN)==LOW) {
        delayMicroseconds(10);
        if(++cnt>MESS_ISR_TIMEOUT)  return;
      }
      if(cnt==0)  return;

      cnt=0;
      while(digitalRead(IR_PIN)==HIGH) {
        delayMicroseconds(10);
        if(++cnt>MESS_ISR_TIMEOUT)  return;
      }
    }
  }
}

/* ------------------------------------------------------------------------------------
  Holt fünf Tasten, anhand derer die Signal-Zeiten, Signal-Häufigkeiten und die
  Datenflanke ermittelt werden.
  Liefert 0 bei Erfolg und 1 bei Fehler.
*/
unsigned char ZeitenEinmessen(void)
{
  unsigned char proben, i, j;

  for(j=0; j<2; j++) {
    ird.rs.m1.signalZeiten_idx[j]=0;
    for(i=0; i<10; i++) {
      ird.rs.m1.signalZeiten[j][i]=0;      // Debug-Kosmetik
      ird.rs.m1.signalZeitenMin[j][i]=0;   // Debug-Kosmetik
      ird.rs.m1.signalZeitenMax[j][i]=0;   // Debug-Kosmetik
      ird.rs.m1.signalZeitenAnzahlen[j][i]=0;
    }
  }
  zeichenVerfuegbar_g=0;
  isr_modus=ISR_MODE_EINMESSUNG;

  MeldungAusgeben(10); // Bitte jetzt fuenf Tasten druecken !
  for(proben=0; proben<5;) {
    if(zeichenVerfuegbar_g==1) {
      zeichenVerfuegbar_g=0;
      proben++;
      gutTon();
//      messung1Anzeigen();
    }
    if((isr_ist_aktiv_g==0) && (zeichenVerfuegbar_g==0)) {
      isr_ist_aktiv_g=1;
      attachInterrupt(0, isr_ir, FALLING);
    }
    delay(10);
  }

  signalZeitenSortieren(FLANKE_2H);
  signalZeitenSortieren(FLANKE_2L);
//  messung1Anzeigen();

  ird.datenFlanke=DatenFlankeFinden();

  if(ird.datenFlanke==255) {
    MeldungAusgeben(130); // Fehler!!! Kann Datenflanke nicht ermitteln!
    return(1);
  }

  ird.signalZeiten_idx=ird.rs.m1.signalZeiten_idx[ird.datenFlanke];
  for(i=0; i<10; i++) {
    ird.signalZeiten[i]=ird.rs.m1.signalZeiten[ird.datenFlanke][i];
  } // ab jetzt kann von ird.rs.m1.xxx auf ird.rs.m2.xxx umgeschaltet werden
  return(0);
}

/* ------------------------------------------------------------------------------------
  Setzt in "ird.rs.m2.maske_syncbits[]" an den Stellen 1'en, an denen bei zwei
  Tastencodes in "ird.rs.m2.signal[0][]" und "ird.rs.m2.signal[1][]" SyncBits (Wert>1)
  gefunden werden.
  Liefert die Anzahl der bereits vorhandenen und passenden SyncBits als Rückgabe.
*/
unsigned char SyncBitMaskeAufbauen(void)
{
  unsigned char i, sbb;

  for(sbb=i=0; i<100; i++) {
    if(BitInSignalTesten(0, i)>1 && BitInSignalTesten(1, i)>1) {
      // wenn in beiden Tastencodes SyncBits sind
      if(BitInMaskeSyncBitsTesten(i)==1) {
        // wenn es zur bereits vorhandenen SyncBit-Maske passt
        sbb++;
      }
      BitInMaskeSyncBitsSetzen(i, 1);
    }
  }
  return(sbb);
}

/* ------------------------------------------------------------------------------------
  Holt solange Tastencodes und bildet dafür die SyncBit-Maske, bis erstens eine
  Maske da ist und zweitens eine weitere Taste auf diese Maske gepasst hat.
*/
void SyncbitsSicherErkennen(void)
{
  unsigned char i, proben, sbb;

  ird.signal_idx=0;
  zeichenVerfuegbar_g=0;

  BufferSchnellLoeschen(ird.rs.m2.signal[0]);
  BufferSchnellLoeschen(ird.rs.m2.signal[1]);
  BufferSchnellLoeschen(ird.rs.m2.maske);
  BufferSchnellLoeschen(ird.rs.m2.maske_syncbits);

  MeldungAusgeben(11); // Bitte jetzt ein paar Tasten lange (ca. eine Sekunde lang) druecken !

  isr_modus=ISR_MODE_CODEMESSUNG;
  for(proben=0; proben<10;) {
    if(zeichenVerfuegbar_g==1) {
      zeichenVerfuegbar_g=0;
      proben++;
      gutTon();

      if(ird.signal_idx==1) { // wenn signal[0] und signal[1] da sind...
        sbb=SyncBitMaskeAufbauen();
//        messung2Anzeigen();
        for(i=0; i<50; i++) { // hochschieben und Platz für nächste Probe schaffen
          ird.rs.m2.signal[0][i]=ird.rs.m2.signal[1][i]; // signal[0][] = signal[1][]
        }
        BufferSchnellLoeschen(ird.rs.m2.signal[1]);      // signal[1] = 0
        if(sbb>0) break;
      } 
      else {      
        ird.signal_idx++;
      }
    }

    if((isr_ist_aktiv_g==0) && (zeichenVerfuegbar_g==0)) {
      isr_ist_aktiv_g=1;
      attachInterrupt(0, isr_ir, FALLING);
    }
    delay(10);
  }
  if(sbb==0) {  // FB hat Probleme mit den SyncBits...
    MeldungAusgeben(200); // Probleme mit der SyncBit-Erkennung!
    ird.signalZeiten_idx=2;  // egal...SyncBits einfach abschneiden
  }
  MeldungAusgeben(201); // Fertig mit der SyncBit-Erkennung!
  schlechtTon();
}

/* ------------------------------------------------------------------------------------
  Holt zwei Codes für eine Taste und prüft auf Gleichheit der 0- und 1-Bits.
  Bei Ungleichheit wird in der entsprechenden Masken-Position eine 2 vermerkt.
  Durch diesen Test können Sonderbits ausgeblendet werden, die bei jedem
  Tastendruck togglen.
*/
void SonderBitsErkennen(void)
{
  unsigned char i, proben, b1, b2;

  ird.signal_idx=0;
  zeichenVerfuegbar_g=0;

  BufferSchnellLoeschen(ird.rs.m2.signal[0]);
  BufferSchnellLoeschen(ird.rs.m2.signal[1]);
  BufferSchnellLoeschen(ird.rs.m2.maske);

  MeldungAusgeben(12); // Bitte jetzt eine Taste zweimal lange (ca. eine Sekunde lang) druecken !

  for(proben=0; proben<2;) {
    if(zeichenVerfuegbar_g==1) {
      zeichenVerfuegbar_g=0;
      proben++;
      gutTon();
      if(ird.signal_idx==1) {
        for(i=0; i<100; i++) {
          b1=BitInSignalTesten(0, i);
          b2=BitInSignalTesten(1, i);
          if(b1!=b2 && b1<2 && b2<2) {
            BitInMaskeSetzen(i, 2);
          }
        }
      } 
      else {      
        ird.signal_idx++;
      }
//      messung2Anzeigen();
    }

    if((isr_ist_aktiv_g==0) && (zeichenVerfuegbar_g==0)) {
      isr_ist_aktiv_g=1;
      attachInterrupt(0, isr_ir, FALLING);
    }
    delay(10);
  }
}

/* ------------------------------------------------------------------------------------
  Liefert 1, wenn das Signal in "ird.rs.m2.signal[idx][]" die SyncBits an der
  richtigen Stelle hat. Sonst wird 0 geliefert.
*/
unsigned char SignalPasstAufSyncBits(unsigned char idx)
{
  unsigned char i, flg;

  flg=1;
  // entweder müssen alle SyncBits aus dem Signal auf die Maske passen...
  for(i=0; i<100; i++) {
    if(BitInSignalTesten(idx, i)>1 && BitInMaskeSyncBitsTesten(i)==0) {
      flg=0;  // Signal==SyncBit && SyncBitMaske==0
    }
  }
  if(flg==1) {  // hat komplett gepasst
    return(1);
  }
  // ...oder alle SyncBits aus der Maske müssen auf das Signal passen
  for(i=0; i<100; i++) {
    if(BitInSignalTesten(idx, i)<=1 && BitInMaskeSyncBitsTesten(i)>0) {
      return(0);  // Signal!=SyncBit && SyncBitMaske==1
    }
  }
  return(1);
}

/* ------------------------------------------------------------------------------------
  Läuft über "ird.rs.m2.signal[0][]" und "ird.rs.m2.signal[1][]" und trägt bei 
  Ungleichheit (und wenn keins von beiden ein SyncBit ist) eine 1 in 
  "ird.rs.m2.maske[]" ein.
*/
void MaskeAufbauen(void)
{
  unsigned char i, b1, b2;

  for(i=0; i<100; i++) {
    b1=BitInSignalTesten(0, i);
    b2=BitInSignalTesten(1, i);
    if(b1!=b2 && b1<2 && b2<2) {
      BitInMaskeSetzen(i, 1);
    }
  }
}

/* ------------------------------------------------------------------------------------
  Liest die Tastencodes von allen Tasten und markiert in "ird.rs.m2.maske[]" jeweils
  die Positionen, an denen die Bits unterschiedliche Werte haben.
*/
void DatenFuerCodeanalyseHolen(void)
{
  unsigned char i, proben;
  unsigned int to;

  ird.signal_idx=0;
  ird.anzahl_tasten=0;
  zeichenVerfuegbar_g=0;

  BufferSchnellLoeschen(ird.rs.m2.signal[0]);
  BufferSchnellLoeschen(ird.rs.m2.signal[1]);

  MeldungAusgeben(13); // Bitte jetzt jede vorhandene Taste genau einmal druecken ! Ende durch Pause von 10 Sekunden.

  to=0;
  for(proben=0; proben<100;) {
    if(++to>1000) {    // Ende, wenn 10 Sekunden keine Taste gedrückt wurde
      break;
    }
    if(zeichenVerfuegbar_g==1) {
      zeichenVerfuegbar_g=0;
      proben++;
      to=0;
      if(SignalPasstAufSyncBits(ird.signal_idx)==0) {
        schlechtTon();
//        messung2Anzeigen();
        for(i=0; i<50; i++) {
          ird.rs.m2.signal[ird.signal_idx][i]=0;
        }
      } 
      else {
        gutTon();
        ird.anzahl_tasten++;
//        messung2Anzeigen();

        if(ird.signal_idx==1) {  // wenn zwei Signale vorliegen
          MaskeAufbauen();

          for(i=0; i<50; i++) {  // umkopieren ([0]=[1] und [1]=0)
            ird.rs.m2.signal[0][i]=ird.rs.m2.signal[1][i];
            ird.rs.m2.signal[1][i]=0;
          }
        } 
        else {      
          ird.signal_idx++;
        }
      }
    }

    if((isr_ist_aktiv_g==0) && (zeichenVerfuegbar_g==0)) {
      isr_ist_aktiv_g=1;
      attachInterrupt(0, isr_ir, FALLING);
    }
    delay(10);
  }
  MeldungAusgeben(202); // Fertig mit der Codeanalyse-Vorbereitung!
}

/* ------------------------------------------------------------------------------------
  Sucht in "ird.rs.m2.maske[]" ab "startpos" eine Folge von 1-Bits der Mindestlänge
  "min_bits". Wenn so eine Folge gefunden wurde, werden "start_bitpos" und "stop_bitpos"
  auf die entsprechenden Grenz-Positionen gesetzt und der Wert 1 zurückgeliefert.
  Wenn keine solche Folge gefunden wurde, wird der Wert 0 zurückgeliefert.
*/
unsigned char DatenBereichsGrenzenSuchen(unsigned char startpos, unsigned char min_bits, unsigned char *start_bitpos, unsigned char *stop_bitpos)
{
  unsigned char i;

  *start_bitpos=0;
  *stop_bitpos=0;
  i=startpos;
  min_bits--;  // (Stop-Start) also z.B. 21-17=4 - obwohl es 5 Bit lang ist

  while(((*stop_bitpos-*start_bitpos)<min_bits) && i<100) {
    for(; i<100; i++) {  // erste 1 anfahren
      if(BitInMaskeTesten(i)==1) {
        break;
      }
    }
    *start_bitpos=i;
  
    for(; i<100; i++) {  // folgende !1 anfahren
      if(BitInMaskeTesten(i)!=1) {
        break;
      }
    }
    *stop_bitpos=i-1;
  }
  if(*start_bitpos==100) return(0);
  return(1);
}

/* ------------------------------------------------------------------------------------
  Liefert aus den 50-Byte-Buffer "buf" das Bitmuster zwischen "start_bitpos" und
  "stop_bitpos". Es werden 0'en und 1'en ins Ergebnis kopiert, >1 wird zu 0.
  "start_bitpos" bis "stop_bitpos" sollte <=16 sein - sonst kommt Quatsch.
*/
unsigned int BitfolgeAusBuffer(unsigned char *buf, unsigned char start_bitpos, unsigned char stop_bitpos)
{
  unsigned char i;
  unsigned int rc;

  for(rc=0, i=start_bitpos; i<=stop_bitpos; i++) {
    rc*=2;
    if((i&1)==0) { if(((buf[i>>1])&0x0f)==0x01) rc++; }
    else         { if(((buf[i>>1])&0xf0)==0x10) rc++; }
  }
  return(rc);
}

/* ------------------------------------------------------------------------------------
  Sucht in "ird.rs.m2.maske[]" nach 1'er-Folgen, die mindestens so lang sind, dass
  für alle Tasten ein eigener Code möglich ist.
  Weiterhin wird geprüft, ob eine zweite gleich lange 1'er-Folge existiert, in der
  das Bitmuster in "ird.rs.m2.signal[0][]" wiederholt oder invertiert vorkommt. Wenn
  dem so ist, wird der Zwischenraum zwischen den beiden 1'er-Folgen geprüft, ob dort
  1'en vorkommen. Wenn ja, werden diese zur 1'er-Folge hinzugefügt (um so auch
  unterbrochene Bitmuster verarbeiten zu können - also Maske z.B. = 1111101), sofern
  sie mit nur einem 0-Bit getrennt vor oder hinter den 1'er Folgen auftreten. In so
  einem Fall wird der Wert 2 zurückgegeben und die Funktion ist erneut aufzurufen.
  Existiert keine zweite gleich lange Folge, wird geprüft, ob die eine Folge ggf.
  halbiert werden kann. Sie kann halbiert werden, wenn sie in zwei gleich lange
  Teile geteilt werden kann und im zweiten Teil eine Wiederholung oder Invertierung
  des ersten Teils in "ird.rs.m2.signal[0][]" erkannt wird.

  Rückgabewerte:
    0 : Wert gefunden
    1 : Keinen passenden Bereich gefunden
    2 : Änderungen an Maske durchgeführt. Nochmal aufrufen
*/
unsigned char DatenBereichsGrenzenFinden(void)
{
  unsigned char i, rc, min_bits;
  unsigned char start_bitpos[2], stop_bitpos[2];
  unsigned int code[3], zwischenraum;

  min_bits=benoetigteBits(ird.anzahl_tasten);

  if(DatenBereichsGrenzenSuchen(0, min_bits, &start_bitpos[0], &stop_bitpos[0])==0) {
    // wenn keine hinreichend lange Folge gefunden wird, zwei Bit abziehen, um sie
    // dann später hoffentlich via 10xxx oder xxx01 wieder anhängen zu können.
    min_bits-=2;
  }

  if(DatenBereichsGrenzenSuchen(0, min_bits, &start_bitpos[0], &stop_bitpos[0])==1) {
    // ersten passend großen Bereich gefunden
    if(DatenBereichsGrenzenSuchen(stop_bitpos[0]+1, min_bits, &start_bitpos[1], &stop_bitpos[1])==1) {
      // zweiten passend großen Bereich dahinter gefunden
      if((stop_bitpos[0]-start_bitpos[0]) == (stop_bitpos[1]-start_bitpos[1])) {
        // beide Bereiche sind gleich lang
        code[0]=BitfolgeAusBuffer(ird.rs.m2.signal[0], start_bitpos[0], stop_bitpos[0]);
        code[1]=BitfolgeAusBuffer(ird.rs.m2.signal[0], start_bitpos[1], stop_bitpos[1]);
        code[2]=code[1]^BitfolgeAusBuffer(ird.rs.m2.maske, start_bitpos[0], stop_bitpos[0]);
        if((code[0]==code[1]) || (code[0]==code[2])) {
          // zweiter Code ist Wiederholung oder Invertierung des ersten Codes
          // noch Grenzen testen
          zwischenraum=BitfolgeAusBuffer(ird.rs.m2.maske, stop_bitpos[0]+1, start_bitpos[1]-1);
          if(zwischenraum==0) {
            // nur statische Bits zwischen den Folgen xxxxx und yyyyy
            // es gibt also weder 10xxxxx-----10yyyyy
            //               noch   xxxxx01-----yyyyy01
            // Ausstieg mit den Daten für die erste Folge
          } else if(BitInMaskeTesten(start_bitpos[0]-1)==0 && BitInMaskeTesten(start_bitpos[0]-2)==1) {
            // "10" vor der ersten Folge
            if(BitInMaskeTesten(start_bitpos[1]-1)==0 && BitInMaskeTesten(start_bitpos[1]-2)==1) {
              // "10" vor der zweiten Folge
              BitInMaskeSetzen(start_bitpos[0]-1, 1);  // "10" zu "11" umwandeln...
              BitInMaskeSetzen(start_bitpos[1]-1, 1);
              return(2);                               // ...und Test neu aufsetzen
            } else {
              // beim Zweiten passt es nicht
              // Ausstieg mit den Daten für die erste Folge
            }
          } else if(BitInMaskeTesten(stop_bitpos[0]+1)==0 && BitInMaskeTesten(stop_bitpos[0]+2)==1) {
            // "01" hinter der ersten Folge
            if(BitInMaskeTesten(stop_bitpos[1]+1)==0 && BitInMaskeTesten(stop_bitpos[1]+2)==1) {
              // "01" hinter der zweiten Folge
              BitInMaskeSetzen(stop_bitpos[0]+1, 1);  // "01" zu "11" umwandeln...
              BitInMaskeSetzen(stop_bitpos[1]+1, 1);
              return(2);                              // ...und Test neu aufsetzen
            } else {
              // beim Zweiten passt es nicht
              // Ausstieg mit den Daten für die erste Folge
            }
          } else {
            // passt nicht aufs 01-XXX- oder XXX-01-Schema
            // Ausstieg mit den Daten für die erste Folge
          }
        } else {
          // zweite Folge ist weder Wiederholung noch Invertierung
          // Ausstieg mit den Daten für die erste Folge
        }
      } else {
        // Bereiche nicht gleich groß
        // Ausstieg mit den Daten für die erste Folge
      }
    } else {
      // keinen zweiten passend großen Bereich dahinter gefunden
      if(((stop_bitpos[0]-start_bitpos[0])+1)>=(min_bits*2)) {
        // die Folge ist lang genug, um zweimal die Tastencodes aufnehmen zu können
        if((((stop_bitpos[0]-start_bitpos[0])+1)&1)==0) {
          // wenn gerade Bitanzahl -> halbieren
          stop_bitpos[1]=stop_bitpos[0];
          stop_bitpos[0]=(((stop_bitpos[0]-start_bitpos[0])+1)/2)+start_bitpos[0]-1;
          start_bitpos[1]=stop_bitpos[0]+1;
          code[0]=BitfolgeAusBuffer(ird.rs.m2.signal[0], start_bitpos[0], stop_bitpos[0]);
          code[1]=BitfolgeAusBuffer(ird.rs.m2.signal[0], start_bitpos[1], stop_bitpos[1]);
          code[2]=code[1]^BitfolgeAusBuffer(ird.rs.m2.maske, start_bitpos[0], stop_bitpos[0]);
          if((code[0]==code[1]) || (code[0]==code[2])) {
            // zweiter Code ist Wiederholung oder Invertierung des ersten Codes
            // Ausstieg mit den Daten für die erste Folge
          } else {
            // Codes sind weder Wiederholung noch Invertierung
            stop_bitpos[0]=stop_bitpos[1];
            // Ausstieg mit den ursprünglichen Daten für die erste Folge
          }
        } else {
          // ungerade Bitanzahl
          // Ausstieg mit den Daten für die erste Folge
        }
      } else {
        // die Folge ist nicht lang genug, um zweimal die Tastencodes aufnehmen zu können
        // Ausstieg mit den Daten für die erste Folge
      }
    }
  } else {
    // gar keinen passend grossen Bereich gefunden
    return(1);
  }

  ird.start_bitpos=start_bitpos[0];  // Grenzen von der ersten Folge setzen...
  ird.stop_bitpos=stop_bitpos[0];
  return(0);                         // ...und ferig melden
}

/* ------------------------------------------------------------------------------------
  Zählt in "ird.rs.m2.maske[]" die das Signal einleitenden 0-Bits und legt das
  entsprechende Bitmuster in "ird.signal_prefix" und dessen Länge (max. 16) in
  "ird.signal_prefix_lng" ab.
  Das Bitmuster wird links ausgerichtet - das erste Bit steht also auf 0x8000.
*/
void SignalPrefixFinden(void)
{
  unsigned char i, flg;

  ird.signal_prefix=0;
  ird.signal_prefix_lng=0;
  for(i=flg=0; i<16; i++) {  // maximal Wortlänge
    if(flg==0 && BitInMaskeTesten(i)==0) {
      ird.signal_prefix_lng++;
      ird.signal_prefix*=2;
      if(BitInSignalTesten(0, i)>0) {
        ird.signal_prefix++;
      }
    } else {
      flg=1;    // Flag für "1 gefunden" setzen
      ird.signal_prefix*=2;  // oben ausrichten
    }
  }
}

/* ------------------------------------------------------------------------------------
  Liefert eine 1, wenn das Signal in "ird.rs.m2.signal[idx][]" auf das Bitmuster
  in "ird.signal_prefix" passt.
*/
unsigned char SignalPasstAufPrefix(unsigned char idx)
{
  unsigned int i, m;

  for(i=0, m=0x8000; i<ird.signal_prefix_lng; i++, m>>=1) {
    if(BitInSignalTesten(idx, i)>0) {
      if((ird.signal_prefix&m)==0) return(0);  // signal>0 && prefix==0
    } else {
      if((ird.signal_prefix&m)>0)  return(0);  // signal==0 && prefix>0
    }
  }
  return(1);
}

/* ------------------------------------------------------------------------------------
  Liefert aus "ird.rs.m2.signal[0][]" den Tastencode gemäß "ird.start_bitpos" und
  "ird.stop_bitpos".
*/
unsigned int TastenCodeAusSignal(void)
{
  unsigned char i;
  unsigned int rc=0;

  for(i=ird.start_bitpos; i<=ird.stop_bitpos; i++) {
    rc*=2;
    if(BitInSignalTesten(0, i)>0) {
      rc++;
    }
  }
  return(rc);
}


/* ------------------------------------------------------------------------------------
  Setzt globale Variablen auf Initialwerte.
*/
void FernbedienungInit(void)
{
  zeichenVerfuegbar_g=0;
  isr_ist_aktiv_g=0;
  verfuegbaresZeichenHatMainloopDurchlaufen_g=FALSE;
}

/* ------------------------------------------------------------------------------------
  Fragt Tasten zur Einmessung der Fernbedienung ab und setzt "FB_parameter"
  entsprechend. Liefert 0 bei Erfolg, ansonsten >0.
*/
unsigned char FernbedienungVermessen(void)
{
  unsigned char rc;

  if((rc=ZeitenEinmessen())==0) {  // setzt ird.signalZeiten[], ird.signalZeiten_idx, ird.datenFlanke
    FB_parameter.isr_laengeNullBit=ird.signalZeiten[0];
    FB_parameter.isr_flanke=ird.datenFlanke;
  
    SyncbitsSicherErkennen();      // setzt ird.rs.m2.maske_syncbits[]
    SonderBitsErkennen();          // setzt ird.rs.m2.maske[]
    DatenFuerCodeanalyseHolen();   // setzt ird.rs.m2.maske[], ird.anzahl_tasten, nutzt SignalPasstAufSyncBits()
    while((rc=DatenBereichsGrenzenFinden())==2);  // setzt ird.start_bitpos, ird.stop_bitpos
    FB_parameter.isr_startbit=ird.start_bitpos;
    FB_parameter.isr_stopbit=ird.stop_bitpos;
  
    SignalPrefixFinden();          // setzt ird.signal_prefix, ird.signal_prefix_lng
    FB_parameter.isr_prefix=ird.signal_prefix;
    FB_parameter.isr_prefix_bits=ird.signal_prefix_lng;
  }

  detachInterrupt(0);
  zeichenVerfuegbar_g=0;
  isr_ist_aktiv_g=0;
  return(rc);
}

/* ------------------------------------------------------------------------------------
  Die ISR zur Abfrage der Fernbedienung, nachdem "FB_parameter" legale Werte enthält.
  - deaktiviert die ISR (bzw. sich selbst) und setzt "isr_ist_aktiv_g" =0
  - setzt "zeichenVerfuegbar_g" =1, wenn legaler Tastencode empfangen wurde
  - setzt "isr_tastencode" auf den empfangenen Tastencode
*/
void Fernbedienung_isr(void)  // wird bei IR_PIN=="HIGH->LOW" aufgerufen
{
  unsigned char i, bitnr;
  unsigned int cnt, cnt1, cnt2, db, prefix;

  if(digitalRead(IR_PIN)==HIGH) return;
  detachInterrupt(0);
  isr_ist_aktiv_g=isr_tastencode=bitnr=prefix=0;

  for(i=0; i<100; i++) {
    for(cnt1=0; digitalRead(IR_PIN)==LOW  && cnt1<15000; cnt1++) delayMicroseconds(10);
    for(cnt2=0; digitalRead(IR_PIN)==HIGH && cnt2<15000; cnt2++) delayMicroseconds(10);
    if(FB_parameter.isr_flanke==1) cnt=cnt2;
    else                           cnt=cnt1;
    if((cnt/20)>15) db=cnt/20; else db=15;
    if(bitnr<16) prefix*=2;  // vom Prefix werden maximal 16 Bit geprüft
    if(bitnr>=FB_parameter.isr_startbit && bitnr<=FB_parameter.isr_stopbit) isr_tastencode*=2;
    if(!(cnt>(FB_parameter.isr_laengeNullBit-db) && cnt<(FB_parameter.isr_laengeNullBit+db))) {
      if(bitnr>=FB_parameter.isr_startbit && bitnr<=FB_parameter.isr_stopbit) isr_tastencode++;
      else if(bitnr<FB_parameter.isr_prefix_bits) prefix++;
    }
    bitnr++;
    if(cnt>=15000) break;
  }
  if(prefix==FB_parameter.isr_prefix) zeichenVerfuegbar_g=1;

  if(cnt<15000) { // Rest ggf. weglesen
    while(1) {    // Ende über return
      for(cnt=0; digitalRead(IR_PIN)==LOW  && cnt<15000; cnt++) delayMicroseconds(10);
      if(cnt==0) return;
      for(cnt=0; digitalRead(IR_PIN)==HIGH && cnt<15000; cnt++) delayMicroseconds(10);
    }
  }
}

/* ------------------------------------------------------------------------------------
  Holt eine Taste von der Fernbedienung. Es wird der Tastenkode zurückgeliefert.
*/
unsigned int warteAufTaste(void)
{
  while(1) {
    if(zeichenVerfuegbar_g==1) {
      beep(1000, 100);
      zeichenVerfuegbar_g=0;
      return(isr_tastencode);
    }

    if((isr_ist_aktiv_g==0) && (zeichenVerfuegbar_g==0)) {
      isr_ist_aktiv_g=1;
      attachInterrupt(0, Fernbedienung_isr, FALLING);
    }
    delay(10);
  }
}

/* ------------------------------------------------------------------------------------
  Fragt ANZAHL_TASTEN Tasten ab und schreibt die jew. Kodes in "tastenfunktion[]".
*/
void TastencodesZuweisen(void)
{
  unsigned char i;

  for(i=0; i<ANZAHL_TASTEN; i++) {
    lcd.clear();         lcd.print("nun Taste fuer:");
    lcd.setCursor(1, 1); lcd.print(tastennamen[i]);
    tastenfunktion[i]=warteAufTaste();
  }
}

/* ------------------------------------------------------------------------------------
  Liefert den TastenIndex zu "t" ( 0 - ANZAHL_TASTEN-1 ).
  Ist "t" kein legaler TastenIndex, wird ANZAHL_TASTEN (für "unbekannt") geliefert.
*/
unsigned char tastenIndex(unsigned int t)
{
  unsigned char i;

  for(i=0; i<ANZAHL_TASTEN; i++)
    if(t==tastenfunktion[i]) break;
  return(i);
}



/* ---------------------------------------------------------------
  Liefert die Tastengruppe zu "c" als Return-Wert und ändert "k"
  auf den zu "c" passenden IRKey-Wert
*/
unsigned int istAktiveTaste(int c, unsigned char *k)
{
  unsigned int grp;

  switch (c) {  
    case 0:  grp=IRgrp_ZIFFERN;  *k=0;            break;
    case 1:  grp=IRgrp_ZIFFERN;  *k=1;            break;
    case 2:  grp=IRgrp_ZIFFERN;  *k=2;            break;
    case 3:  grp=IRgrp_ZIFFERN;  *k=3;            break;
    case 4:  grp=IRgrp_ZIFFERN;  *k=4;            break;
    case 5:  grp=IRgrp_ZIFFERN;  *k=5;            break;
    case 6:  grp=IRgrp_ZIFFERN;  *k=6;            break;
    case 7:  grp=IRgrp_ZIFFERN;  *k=7;            break;
    case 8:  grp=IRgrp_ZIFFERN;  *k=8;            break;
    case 9:  grp=IRgrp_ZIFFERN;  *k=9;            break;
    case 10: grp=IRgrp_ONOFF;    *k=IRkey_ONOFF;  break;  // ON/OFF
    case 11: grp=IRgrp_CSRLTRT;  *k=IRkey_CSRLT;  break;  // CSR LT
    case 12: grp=IRgrp_CSRLTRT;  *k=IRkey_CSRRT;  break;  // CSR RT
    case 13: grp=IRgrp_CSRUPDN;  *k=IRkey_CSRUP;  break;  // CSR UP
    case 14: grp=IRgrp_CSRUPDN;  *k=IRkey_CSRDN;  break;  // CSR DN
    case 15: grp=IRgrp_MENU;     *k=IRkey_MENU;   break;  // MENU
    case 16: grp=IRgrp_ENTER;    *k=IRkey_ENTER;  break;  // ENTER
    case 17: grp=IRgrp_EXIT;     *k=IRkey_EXIT;   break;  // EXIT
    case 18: grp=IRgrp_PRGUPDN;  *k=IRkey_PRGUP;  break;  // PRG UP
    case 19: grp=IRgrp_PRGUPDN;  *k=IRkey_PRGDN;  break;  // PRG DN
    case 20: grp=IRgrp_SLEEP;    *k=IRkey_SLEEP;  break;  // SLEEP
    default: grp=IRgrp_ILLEGAL;  *k=99;
  }
  return(grp);
}

/* ---------------------------------------------------------------
  Liefert TRUE, wenn ein im Kontext legales Zeichen seit Beginn der 
  Main-Loop verfügbar ist.
  "aktuelle_Tastengruppe_g" wird auf die jeweilige Tastengruppe gesetzt.
  "aktuelle_Taste_g" wird auf den jeweiligen Tastencode gesetzt.
*/
unsigned char legalesZeichenVerfuegbar(unsigned int aktive_tasten)
{
  if(zeichenVerfuegbar_g==TRUE && verfuegbaresZeichenHatMainloopDurchlaufen_g==TRUE) {
    aktuelle_Tastengruppe_g=istAktiveTaste(tastenIndex(isr_tastencode), &aktuelle_Taste_g);
    if((aktuelle_Tastengruppe_g & aktive_tasten)>0) {
      return(TRUE);
    }
  }
  return(FALSE);
}

