#include <LiquidCrystal.h>
#include "screen.h"

char scrbuf[SCR_ROWS][SCR_COLS];  // Vorbereitungs-Schirm
char scrout[SCR_ROWS][SCR_COLS];  // Anzeige-Schirm
int scrbuf_x, scrbuf_y;         // aktuelle Cursor-Position

//                rs, enable, d4, d5, d6, d7
LiquidCrystal lcd(8,  7,      3,  4,  5,  6);

/* ---------------------------------------------------------------
  LCD-Ansteuerung über Buffer, so dass das LCD nur dann beschrieben
  wird, wenn sich auch etwas geändert hat
*/
void scrInit(void)
{
  lcd.begin(SCR_COLS, SCR_ROWS);
}

void scrUpdateDisplay(void)
{
  char *ptr_src, *ptr_dst;
  int x, y;

  ptr_dst=scrout[0];
  ptr_src=scrbuf[0];
  for(y=0; y<SCR_ROWS; y++) {
    for(x=0; x<SCR_COLS; x++, ptr_src++, ptr_dst++) {
      if(*ptr_src!=*ptr_dst) {
        lcd.setCursor(x, y);
        lcd.write(*ptr_src);
        *ptr_dst=*ptr_src;
      }    
    }
  }
}

void scrPrint(char *ptr_src)
{
  char *ptr_dst;

  ptr_dst=scrbuf[0]+scrbuf_y*SCR_COLS+scrbuf_x;
  for(; *ptr_src!='\0'; ptr_src++, ptr_dst++) {
    *ptr_dst=*ptr_src;
    scrbuf_x++;
  }
}

void scrPrint(int x, int y, char *ptr_src)
{
  scrSetCursor(x, y);
  scrPrint(ptr_src);
}

void scrPrint(int data)
{
  char buf[SCR_COLS];

  scrPrint(itoa(data, buf, 10));
}

void scrPrint(int data, int stellen)
{
  char buf[SCR_COLS]="000";
  int l;

  itoa(data, buf+3, 10);
  l=strlen(buf);
  scrPrint(buf+(l-stellen));
}

void scrPrint(int x, int y, int data, int stellen)
{
  scrSetCursor(x, y);
  scrPrint(data, stellen);
}

void scrCls(void)
{
  char (*ptr1)=scrbuf[0];
  char (*ptr2)=scrout[0];
  int i;

  scrbuf_x=scrbuf_y=0;
  for(i=0; i<(SCR_COLS*SCR_ROWS); i++, ptr1++, ptr2++) {
    *ptr1=' ';  *ptr2='x';
  }
  scrNoCursorBlink();
  scrUpdateDisplay();
}

void scrSetCursor(int x, int y)
{
  scrbuf_x=x;  // von 0 bis SCR_COLS-1
  scrbuf_y=y;  // von 0 bis SCR_ROWS-1
}

void scrCursorBlink(void)
{
  lcd.setCursor(scrbuf_x, scrbuf_y);
  lcd.blink();
}

void scrNoCursorBlink(void)
{
  lcd.noBlink();
}

